/* 2 jan 2000. Daniel Hellerstein (danielh@crosslink.net)
  UpDater: An rsync based server-to-server updater.
        This is the "central" component.
 For details, see updater.doc
*/

/**** User changeable parameters ****/

/* 1 to use K and M abbreviations when  displaying filesizes, 0 to show full number */
use_abbrev=1 

/* Always use cache entry if it's current. That is, suppress NOCACHE=1 option */
always_use_cache=1

/* add crc's of file in package list (required for "crc32 check" by remote
   servers. 1=yes, 0=no */
addcrc=1

/* colors and or background used in some html displays */
back1='bgcolor="#88aa66"'

/* allow use of GZIP encoding */
candogzip=1

/* allow use of Rsync-differencing encoding */
candorsync=1

/* set of "encryption" specs. ENCRYPT.!BLOWFISH is an example,
  you can delete it if you will never use blowfish */
ENCRYPT.!blowfish='bf -q -e %in %out "%key"'

/* to NOT send the "list of files" as they are available, set this to
   1.  Sending "as available" helps prevent timeouts */
not_send_pieces=0

/* list of files, in or under the updater directory, that the SEND= option
   is allowed to "send to the client" */
send_files='UPDATER.DOC RXGZLIB.DLL GZIP.EXE RXRSYNC.DLL UPCLIENT.CMD UPDATEME.CMD'

/*  the "updater" directory -- should be a fully qualified directory.
    Or, leave it blank to use the default (the UPDATER subdirectory of 
    the GoServe working directory). */

UPDATER_DIR=''
 
 
/**** End of User changeable parameters ****/

updater:    

parse arg  ddir, tempfile, reqstrg,list,verb ,uri,user, ,
          basedir ,workdir,privset,enmadd,transaction,verbose, ,
         servername,host_nickname,homedir,aparam,semqueue,prog_file,reqnum

if verb="" then do
   say " This SRE-http procedure is not meant to be run in stand-alone mode. "
   exit
end  /* Do */
signal on error name badguys; signal on syntax name badguys
crlf = '0d0a'x

/* may need this below */
parse var privset privset1 ',' privset2
privset=privset1

/* look in privset2 for ?UPDATER= "Secret privilege" */
shared_secret='UPDATER'
do until privset2=''
   parse upper var privset2 apriv privset2
   if abbrev(apriv,'UPDATER:')=0 then iterate
   parse var apriv . ':' shared_secret
   shared_secret=strip(shared_secret)
   leave
end /* do */

if updater_dir='' then do
   'header add x-updater-warning: No default updater directory'
   updater_dir=strip(basedir,,'\')'\UPDATER'
end /* do */

if dosisdir(updater_dir)=0 then do
     'header add x-updater: eRror'
     'string UPDATER configuration error. No such updater directory: 'updater_dir      
      return ' '
end

package_dir=updater_dir'\PACKAGE'
if dosisdir(package_dir)=0 then do
     'header add x-updater: erRor'
     'string UPDATER configuration error. No such package directory: 'package_dir
      return ' '
end

temp_dir=updater_dir'\TEMP'
if dosisdir(temp_dir)=0 then do
     'header add x-updater: errOr'
     'string UPDATER configuration error. No such temporary directory: 'temp_dir
      return ' '
end

call check_dlls         /* load some dlls */

/* pull out arguments */
args.=''
selects.=''
if list='' then list='package=!list&ashtml=1'
do until list=""                /* get user input */
   parse var list a1 '&' list
   parse var a1 a1a '=' a1b0
   a1b0=packur(translate(a1b0,' ','+'||'0d0a00900'x))
   a1a='!'||translate(strip(a1a))
   args.a1a=strip(a1b0)
   args.!arglist=args.!arglist' 'a1a
end

call do_central

return ' '

/*************************************/
/* here if updater is acting as a central server */
do_central:

anaction=translate(fixit(args.!ACTION))


/* send UPCLIENT.CMD, etc */
if strip(translate(args.!SEND))<>'' then do
 if wordpos(strip(translate(args.!send)),translate(SEND_FILES))=0 then do
       'header add x-updater: erroR'
       'string no such file available 'args.!send
       return ' '
  end 
  afile=updater_dir'\'||strip(args.!send)
  foo=stream(afile,'c','query exists')
  if foo='' then do
     call lineout tempfile,'<html><head><title>UpClient not available</title></head>'
     call lineout tempfile,'<body 'back1'> '
     call lineout tempfile,'<h3>File not available </h3>'
     call lineout tempfile,'Sorry, 'args.!send ' is not available.'
     call lineout tempfile,'</body></html>'
     call lineout tempfile
     status=sref_gos('FILE ERASE type text/html name 'tempfile )
     return status
   end
   else do
     if args.!text=1 then
        status=sref_gos('FILE type text/plain encrypt 0:0 name 'foo,,,'HEADER ADD Content-Disposition: attachment; filename="'args.!send'"' )
     else
        status=sref_gos('FILE type application/octet-stream encrypt 0:0 name 'foo,,,'HEADER ADD Content-Disposition: attachment; filename="'args.!send'"' )
     return status
   end 
end 

/* show some documentation on upclient */
if anaction='UPCLIENTDOC' then do
       call tell_upclient
       FOO=SREF_GOS('FILE erase TYPE text/html NAME 'tempfile)
       return foo
end /* do */

/* Ask REMOTE CLIENT for some options */
if abbrev(anaction,'GETL2')=1 then do    
       call get_client_options
       FOO=SREF_GOS('FILE erase TYPE text/html NAME 'tempfile)
       return foo
end

/* verify that the client's shared secret is correct */
if strip(translate(args.!VERIFY))<>'' then do
    parse upper var args.!verify nonce';'md4 . ; md4=strip(md4)
    tocheck=space(translate(nonce||shared_secret),0)
    md4b=translate(rx_md4(tocheck))
    if  md4=md4b then
       av='VERIFY=1'
    else
       av='VERIFY=0'
    status=sref_gos('VAR type application/x-updater name av ',av)
    return status
end /* do */


/* (the default) return a list of current packages */
IF ARGS.!PACKAGE='?' | translate(strip(ARGS.!PACKAGE))='!LIST' then DO
      call do_packagelist
      'header add X-updater: package_list '
      if astext=1 then
         status=sref_gos('FILE erase type text/plain name 'tempfile )
      else
         status=sref_gos('FILE erase type text/html name 'tempfile )
      return status
end 


package_file=package_dir'\'||strip(args.!package)||'.PKG'
isize=stream(package_file,'c','query size')
if (isize=0 | isize='') then do
    'header add x-updater: error'
    'string  UpDater ERROR: empty or missing package: 'args.!package
    return ' '
end

stat=read_package(package_file,isize)             /* read in package file to package_opts. */

/* see if any privileges required */
if package_opts.!Privs='' | pos('*',package_opts.!Privs)>0 then do
        nop
end 
else do               /* check privs */
      OK=check_privs(privset,package_opts.!privs)
      if ok<>'OK' then return ''   /* check_privs will issue auth request */
end /* do */


/* Otherwise, return info on a selected package */
/*  ? -- a request for listing of currently available packages
    PACKAGE_NAME -- a name of a package (must correspond to a
                    file in the PACKAGE_DIR 
*/




/* check action. */

 select

/* list contents of this package */
    when abbrev(ANACTION,'DISP') then DO                
       call disp_package
       'header add X-updater: display_package '
       FOO=SREF_GOS('FILE erase TYPE text/html NAME 'tempfile)
       return foo
    end    

    when  abbrev(anaction,'GETL3')=1 then do    
      xlist='get=filelist3'
      call do_getl3
      ah='header add X-updater: 'xlist
      FOO=SREF_GOS('FILE erase TYPE text/html NAME 'tempfile,,,ah)
      return foo
   end

    when  abbrev(anaction,'GETL4')=1 then do    /* same as send_get, with extra constraints */
       xlist='get=filelist'
       if candorsync=1 then xlist=xlist||',rsync'
       if candogzip=1 then xlist=xlist||',gzip'
       selects.!ALL=0                           /* selects.!ALL='' means "no selective files" */
       if args.!ALL=1 then do
          selects.!ALL=1
       end /* do */
       else do                  /* create a selection list  */
          do mm=1 to words(args.!ARGLIST)
               aw=strip(word(args.!arglist,mm))
               if abbrev(aw,'!F_')+abbrev(aw,'!DIR_')=0 then iterate
               if abbrev(aw,'!DIR_')=1 then do
                  if strip(args.aw)=1 then do
                     parse var aw . '_' jj .
                     selects.jj=1
                  end
               end
               if abbrev(aw,'!F_')=1 then do
                    parse var aw . '_' entry '_' . 
                    aff=args.aw ; aff2='!'||aff
                    selects.entry.aff2=1
               end 
          end          /* scan of args. */
       end             /* not ALL */

       call send_gets
       if broken=1 then return '0 '
       if astext=1 then do
          if do_multi=0 then do
             ah='header add X-updater: 'xlist
             FOO=SREF_GOS('FILE erase TYPE application/x-updater encrypt 0:0 NAME 'tempfile,,,ah) 
          end
          else do
              foo=sref_multi_send('',,'1E',0)
          end
       end
       else do
          ah='header add X-updater: 'xlist
          FOO=SREF_GOS('FILE erase TYPE text/html NAME 'tempfile,,,ah)
       end
       return foo
   end


/* return a "get list". The remote server should request each file on this
   list, subject to date checking, possibly using rsync */

    when abbrev(anaction,'GETL')=1 then do    /* set package list ! */
       xlist='get=filelist'
       if candorsync=1 then xlist=xlist||',rsync'
       if candogzip=1 then xlist=xlist||',gzip'
       call send_gets
       if broken=1 then return '0 '
       if astext=1 then do
          if do_multi=0 then do
             ah='header add X-updater: 'xlist
             FOO=SREF_GOS('FILE erase TYPE application/x-updater encrypt 0:0 NAME 'tempfile,,,ah) 
          end
          else do
              foo=sref_multi_send('',,'1E',0)
          end
       end
       else do
          ah='header add X-updater: 'xlist
          FOO=SREF_GOS('FILE erase TYPE text/html NAME 'tempfile,,,ah)
       end
       return foo
    end

/* get a file from a package */
    when abbrev(anaction,'FILE')=1 then do   
        call get_file
        return result
    end
    otherwise 
 end  /* select */

'header add x-updater: error'
'string UpDater: No action was specified '

return ' '



/*************************************/
/* get client options */
get_client_options:
packagename=translate(strip(args.!package))
call lineout tempfile,'<html><head><title>Updater Client options </title></head>'
call lineout tempfile,'<body 'back1'>'
CALL LINEOUT TEMPFILE,'<h3>Please select a few UpDater  options</h3>'
call lineout tempfile,'Files listed in the <tt>'args.!package '</tt> package (on <tt> ' servername'</tt>)'
call lineout tempfile,'will be UpDated to your computer.'
   
/* create a form  */
call lineout tempfile,'<form action="/UPDATER" method="GET">'
call lineout tempfile,'<input type="hidden" name="action" value="GETL">'
call lineout tempfile,'<input type="hidden" name="package" value="'args.!package'">'
call lineout tempfile,'<input type="hidden" name="client" value="1">'
if args.!GETUSER=1 then do
  call lineout tempfile,'<input type="hidden" name="GETUSER" value="1">'
end


call lineout tempfile,'<input type="Submit" value="Get the files in 'packagename'... ">'
call lineout tempfile,'<table><tr valign="top">'

if args.!GETUSER=1 then do
  call lineout tempfile,'<td valign="top">Username and password  </td>'
  call lineout tempfile,'<td>Username:<input type="text" name="USER" size=10  > '
  call lineout tempfile,'Password: <input type="password" name="PWD"  size=10> '
  call lineout tempfile,'<br><em>The username and password required for the 'packagename' package may be different'
  call lineout tempfile,' than your 'servername'  "logon" username and password</em>'
end

call lineout tempfile,'</td>'

call lineout tempfile,'<tr valign="top"><td>Destination directory (on your computer)</td> '
call lineout tempfile,'<td><input type="text" name="DESTDIR"  size=30> </td>'

call lineout tempfile,'<tr valign="top"><td>Backup directory (on your computer).</td> '
call lineout tempfile,'<td><input type="text" name="BACKUPDIR" size=30>'
call lineout tempfile,'<br><em> To suppress backing up changed files, leave this field empty.</em></td>'


call lineout tempfile,'<tr valign="top"><td>Update checks &amp; cleanup option</td>' 
call lineout tempfile,'<td><input type="radio" name="CHECK" value="1" checked> '
call lineout tempfile,'Update if newer (using file creation date) <bR>'
call lineout tempfile,'<input type="radio" name="CHECK" value="2" > '
call lineout tempfile,'Update if 32bit (crc32) file hash is different <bR>'
call lineout tempfile,'<input type="radio" name="CHECK" value="3">'
call lineout tempfile,'Update if 128bit (md4) file hash is different <bR>'
call lineout tempfile,'<input type="radio" name="CHECK" value="0">'
call lineout tempfile,'Always update '

call lineout tempfile,'<hr width="25%"> <input type="checkbox" name="DODELETE" value="1" > '
call lineout tempfile,'Delete all existing files (in your destination directory) that are not mentioned in this package'
call lineout tempfile,'</td>'

call lineout tempfile,'<tr><td colspan=2 align="middle"><hr width="50%"><b>Advanced Options </b><br> </td>'
call lineout tempfile,'<tr valign="top"><td valign="top">Range to retrieve</td>'
call lineout tempfile,'<td> Retrieve entries from '
call lineout tempfile,'#<input type="text" name="RANGE1"  size=5></em>'
call lineout tempfile,' to '
call lineout tempfile,'#<input type="text" name="RANGE2"  size=5><em>(leave both boxes blank to retrieve all files)</em>'
call lineout tempfile,'</td>'
call lineout tempfile,'<tr ><td valign="top">Features </td>'
call lineout tempfile,'<td><input type="checkbox" name="nocache" value=1 > Use the latest version of <tt> 'packagename'</tt> ?'
call lineout tempfile,'<br><input type="checkbox" name="dorun" value=1 > Automatically execute <em>installation</em> program <em>(if one is available)</eM> ?'
call lineout tempfile,'<br><input type="checkbox" name="quiet" value=1 > Quiet output (minimal display of status messages)?'
call lineout tempfile,'</td>'
call lineout tempfile,''
call lineout tempfile,''

call lineout tempfile,''
call lineout tempfile,''
call lineout tempfile,'<tr><td colspan=2 align="middle"><hr width="75%"> </td>'

call lineout tempfile,'<tr><td> </td>'
ab='Get the files in 'packagename'...'
call lineout tempfile,'<td align="right"><input type="Submit" value="'||reverse(ab)||'"></td>'

call lineout tempfile,'</table><hr>'

call lineout tempfile,'</form>'

call lineout tempfile,'<p>'
call lineout tempfile,'<a href="/updater?">Return to UpDater main menu</a> (at 'servername')'

call lineout tempfile,'</body></html>'

call lineout tempfile
return 1


/*************************************/
/* get a file from a package
   Uses filename, entry #, and package name to determine 
   actual file.
   Assuming that no_encode does not apply to this file,
      If RSYNC specified, will return an rsync differenc file.
      If Gzip specified, will gzip the file
     (Gzip after RSYNC).
      If encryption specified, encrypt it (after gzip)
      
*/   
get_file:

afile=strip(translate(args.!file,'\','/'),,'\')
aentry=strip(args.!entry)
doerase=' ' ;celist=''
/* say " package_opts.0 " package_opts.0  */
if aentry> package_opts.0 then do
        string "No such entry: "aentry
        exit ' '
end 
if Pos('\',afile)>0 then do
    if package_opts.aentry.!subs<>1 then do
        string "Subdirectories are not accessible "
        exit ' '
    end 
end 

/* append afile to !cfile, but watch out for wildcards */
cfile=package_opts.aentry.!cfile

foo=translate(cfile,' ','\')
aa=word(foo,words(foo))
if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then cfile=cfile'\*.*'

iast=pos('*',cfile)
if iast>0  then do
   islash=lastpos('\',cfile,iast)
   cfile2=left(cfile,islash)||afile
end /* do */
else do
   cfile2=cfile
end /* do */


/* excluded file? */
aa=check_exclude(cfile2)
if aa=1 then do
   foo=sref_response('forbid','Updater can not access 'afile' in package: 'args.!package,servername)  
   return foo
end /* do */
aa=sysfiletree(cfile2,'goo','FT')

if goo.0=0 then do              /* couldn't find it */
   foo=sref_response('notfound','Updater could not find 'afile' in package: 'args.!package,servername)  
   return foo
end /* do */


if anaction='FILE1' then do    /* force without encoding or date checking */
  mtype=sref_MediatypE(cfile2,host_nickname)
  ah='header add content-disposition:  attachment; filename="'||filespec('n',cfile2)||'"'
  foo=sref_gos('file type 'mtype' encrypt 0:0 name 'cfile2,,,ah)
  return foo
end

parse var goo.1 adate orig_size .

xlist='SIZE='||strip(orig_size)||',DIR='||strip(package_opts.aentry.!rdir)

/* note: the date check and the hash check are typically NOT done
  by updater (they are done by rupdater or upclient, using info
  in the package-listing file */

/* check modification date -- no longer used, instead check on client side */
if (args.!mod<>''  ) then do 
  if left(adate,2)<70 then 
    adate='20'||adate
  else
    adate='19'||adate
  if (adate <= args.!mod) then do       /* unmodified */
        'RESPONSE HTTP/1.1 304 Not Modified '
        'Header NoAuto Add Date: '||sref_gmt()
        'Header Add Server: '||extract('serversoftware')' ;'||sref_version(,1)
        'Header add X-UpDater: 'xlist
        'NODATA  '
         return '304 Not Modified'
  end
end

/* check hash? -- no longer used (instead, crc32 check on client side */
if args.!md4<>''  then do
   amd4=strip(translate(rx_md4(cfile2,1)))

   if amd4=strip(translate(args.!Md4)) then do
        'RESPONSE HTTP/1.1 304 Not Modified '
        'Header NoAuto Add Date: '||sref_gmt()
        'Header Add Server: '||extract('serversoftware')' ;'||sref_version(,1)
        'Header add X-UpDater: 'xlist
        'NODATA  '
         return '304 Not Modified'
   end
end /* do */


/* are we allowed to encode this file? */
if strip(args.!No_encoding)=1 then
 noencode=1
else
  noencode=check_noencode(cfile2)


/* rsync this? */
if  noencode=0 & candorsync=1 then do 
  call rsignature
  if result=1 then do   /* found signature */
       dorsync=do_rsync(rheader,cfile2)
       if dorsync<>0 then do
           cfile2=dorsync
           xlist=xlist||',md4='||amd4
           doerase='ERASE '     
           celist='gdiff'
           foo=stream(cfile2,'c','close')
       end /* do */
  end
end

lencfil2=stream(cfile2,'c','query size')
/* gzip this ? */
if args.!gzip=1 & candogzip=1 & noencode=0 & lencfil2>100 then do
  stat=do_gzip(cfile2)
  if stat<>0 then do
      if doerase='ERASE' then do
        foo=sysfiledelete(cfile2)
      end
      doerase='ERASE'
      cfile2=stat
      celist=strip(celist||',gzip',,',')
   end
end

/* encrypt? */
if args.!ENCRYPT<>'' & package_opts.!encrypt<>'' then do
  useenc=''
  aencrypts=translate(strip(translate(args.!encrypt,' ',',')))
  do ll=1 to words(aencrypts)  /* use 1st that matches package_opts.!encrypts */
      a1=strip(word(aencrypts,ll))
      if wordpos(a1,package_opts.!encrypt)>0 then do
         useenc=a1
         leave
      end 
  end
  if useenc='' then return ''  /*  no agreeable encryption, give up */

/*else, encrypt it */
  aa='!'useenc
  encline=encrypt.aa
  cfile2=do_encrypt(cfile2,encline,doerase)
  if cfile2='' then do               /* error, give up */
        if doerase='ERASE' then foo=sysfiledelete(cfile2)
        'RESPONSE HTTP/1.1 500 Internal Server Error '
        'Header NoAuto Add Date: '||sref_gmt()
        'Header Add Server: '||extract('serversoftware')' ;'||sref_version(,1)
        'Header add X-UpDater: Encryption failure using 'useenc
        'NODATA NOWAIT '
         return '500 Internal Server Error'
  end 
  if celist='' then
     celist='ENCRYPT='useenc
  else
     celist=celist',ENCRYPT='useenc
  doerase='ERASE'
end 

ah='header add x-updater: 'xlist
if celist<>'' then do
  ah=ah||'0d0a'x||'header add content-encoding: 'celist
end
foo=sref_gos('file 'doerase ' type application/x-updater  encrypt 0:0 name 'cfile2,,,ah)
return foo

/***********************************/
/* encrypt a file */
do_encrypt:procedure expose  temp_dir shared_secret
parse arg oldfile,doit,doerase

atid=dostid()
apid=dospid()

tmpf=temp_dir'\G'||strip(left(apid,3))||atid||'.ENC'
foo=sysfiledelete(tmpf)

doit=replacestrg(doit,'%IN',oldfile,'ALL')
doit=replacestrg(doit,'%OUT',tmpf,'ALL')
doit=replacestrg(doit,'%KEY',shared_secret,'ALL')
doit=strip(doit)
aa=word(doit,1)

signal on syntax name noenc ; signal on  error name noenc

if pos('(',aa)>0 then do
    doit='foo='doit
    interpret doit 
end
else do
   address cmd doit
end /* do */

signal on error name badguys; signal on syntax name badguys
if doerase='ERASE' then foo=sysfiledelete(oldfile)
doerase='ERASE'

return tmpf             /* the file name to use */

noenc:
signal on error name badguys; signal on syntax name badguys
foo=sysfiledelete(tmpf)

noenc:
return ''             /* failure, send error message (401) */


/***********************************/
/* gzip a file */
do_gzip:procedure expose temp_dir
parse arg afile

atid=dostid()
apid=dospid()

tmpf=temp_dir'\G'||strip(left(apid,3))||atid||'.GZ'
rc=rxgzDeflateFile(afile,tmpf)
if rc=0 then return tmpf
foo=sysfiledelete(tmpf)
return 0

/***************/
/* compute a gdiff difference file, using rsync-signature as a synopsis */
do_rsync:Procedure expose amd4 temp_dir
parse arg rsign,ffile

rsign=pack64(rsign)
blocksize=c2d(substr(rsign,1,4))
nblocks=c2d(substr(rsign,5,1))
synopsis=''
do mm=1 to nblocks
   jat=((mm-1)*8)+6
   foo=substr(rsign,jat,8)
   synopsis=synopsis||left(foo,20)
end

comment='srersync pre-reply '
aa=left(comment,81)||left(blocksize,7)||left(nblocks,9)||left('md4',35)
synopsis=aa||synopsis

atid=dostid()
apid=dospid()
a1=temp_dir||'\'atid'_'apid||'.SYN'
foo=sysfiledelete(a1)
status=rx_rsync_gdiff(ffile,synopsis,a1,1)     /* compute a gdiff, using sypnopsis */
foo=stream(a1,'c','close')
parse var status status amd4 ; status=strip(status)
if status<>'OK' then do
  if verbose>1 then call pmprintf('UpDater rsync error: 'status)
  foo=sysfiledelete(a1)
  return 0              /* error, could not do rsync */
end

foo=stream(a1,'c','open read')
if abbrev(strip(translate(foo)),'READY')=0 then do
  foo=stream(a1,'c','close')
  foo=sysfiledelete(a1)
  if verbose>1 then  call pmprintf('UpDater WARNING: unable to open difference file 'a1)
  return 0
end

issize=stream(a1,'c','query size')
if (issize=0 | issize='') then do       /* some kind of error */
  foo=stream(a1,'c','close')
  foo=sysfiledelete(a1)
  return 0
end

return a1               /* return file containing rsync result */


/****************/
/*Look for a rsync-signature request header. If none, do nothing */
rsignature:
rheader=reqfield('rsync-signature_updater')
if rheader='' then return 0

/* reqfield has problems, so re-read using entire head */
'read header var hh '
do until hh =''
   parse var hh aline '0d0a'x hh
   aline=strip(aline)
   if abbrev(translate(aline),'RSYNC-SIGNATURE_UPDATER:')=1 then do  
       parse var aline . ':' rheader 

       DO UNTIL HH=''
          PARSE VAR hh aline '0d0a'x hh
          if pos(':',aline)>0 then leave
          rheader=rheader||aline
       end              /* do concatenate lines (goserve screwiness correction */
       rheader=strip(rheader)
       leave
   end /* found a rsync-signature set of lines (in header) */
end /* do */
return 1


/*************************************/
/* create a "get list". This list will be used by remote server, 
or remote client, or a remote standalone client.
The remote app should ask for each of these files; possibly checking
datestamps, etc first, and perhaps using rsync & gzip  */

Also, possibly use cache file, or save as cache file.

See updater.doc for details on the structure of these "list of files"

*/
send_gets:                              /* return # of entries written */

/* find all files matching each  .!cfile entry */

/* either send as plain text (for remote server processing), or
  as an html list (for browser-based human processing */
astext=1
iserror=0
broken=0
do_multi=0

/* if remote client (via updater), check for required fields first */
if args.!client=1 then do
  ddir=fixit(args.!destdir)
  if ddir='' then do
    call lineout tempfile,'<html><head><title>ERROR: no destination directory </title></head>'
    call lineout tempfile,'<body 'back1'> '
    call lineout tempfile,'<h3>ERROR: no destination directory </h3>'
    call lineout tempfile,'You forgot to specify a destination directory.'
    call lineout tempfile,'The destination directory is a directory on <b>your</b> computer,'
    call lineout tempfile,'it is where UpDater will place the files in this package.'
    call lineout tempfile,'<p><em>Please use your browser''s BACK key, and select a destination directory</em>'
    call lineout tempfile,'</body></html>'
    call lineout tempfile
    astext=0
    return 1
  end
  if args.!GETUSER=1  then do
      auser=fixit(args.!user) 
      apwd=fixit(args.!pwd) 

      if auser='' | apwd='' then do
        call lineout tempfile,'<html><head><title>ERROR: no username/password </title></head>'
        call lineout tempfile,'<body 'back1'> '
        call lineout tempfile,'<h3>ERROR: no username or no password </h3>'
        call lineout tempfile,'You forgot to specify a username and password.'
        call lineout tempfile,'To UpDate files in this package, you must have an appropriate <b>username and password</b>.'
        call lineout tempfile,'This <em>package specific</em> username and password may, or may <b>not</b>, be the same as the <em>logon username and password</em> required by this server.'
        call lineout tempfile,'<p><em>Please use your browser''s BACK key, and select a username and password</em>'
        call lineout tempfile,'</body></html>'
        call lineout tempfile
        astext=0
        return 1
      end /
  end /* do */
end /* do */

numeric digits 11
nowtime=date('b')+(time('m')/1440)    /* current date/time  -- 1 minute resolution */
parse var reqstrg areq '?' .

/* possibly return a cached filelist */
do yi=1 to 1
   if selects.!ALL<>'' then leave       /* if custom selection, then can't use cache */
   if strip(args.!ashtml)=1 | package_opts.!cache=0 then leave
   if (always_use_cache<>1 & args.!nocache=1) | package_opts.!cachestamp=0 then leave

   if package_opts.!cachestamp+package_opts.!cache<nowtime then leave  

/* IF here, use cached entry! */
   goo=stream(package_opts.!cachefile,'c','open read')
   if abbrev(strip(translate(goo)),'READY')=0 then leave
   sz3=stream(package_opts.!cachefile,'c','query size')
   if sz3=0 | sz3='' then do
      if verbose>1 then call pmprintf(' UpDater warning: empty cache file: 'package_opts.!cachefile)
      leave
   end
   aax=charin(package_opts.!cachefile,1,sz3)
   foo=stream(package_opts.!cachefile,'c','close')
   parse var aax hdrline '0d0a'x aax
   if length(aax)=0 then leave
   aline='Package='args.!package
   aline=aline||' '||'Server='servername
   if package_opts.!encrypt<>"" then aline=aline' Encrypt='||translate(package_opts.!encrypt,',',' ')
   
   CC=GET_CREATED(HDRLINE)
   parse var cc cc1 cc2

   tnow=make_timestamp(package_file)

   if tnow>cc2 then leave         /* if package file has changed, don't use cache */
   aline=aline||' '||'CREATED='CC1
   aline=aline||' '||'pKG_DATE='CC2
   if package_opts.!Privs<>'' & pos('*',package_opts.!Privs)=0 then 
          aline=aline||' NEEDPRIVS=1'
   if candogzip=1 then aline=aline' GZIP=1'
   if candorsync=1 then aline=aline' RSYNC=1'
   aline=aline' CURRENT_TIME='nowtime
   aline=aline' SELECTOR='areq

   if args.!client=1 then call add_remote_client_stuff

   call charout tempfile,aline||'0d0a'x
   call charout tempfile,aax
   if verbose>2 then say "UpDater: using cache " package_opts.!cachefile
   foo=stream(tempfile,'c','close')
   return 1
end

cachefile=''            /* '' means "do not save to cache file */

/* NOT from cache... */
if strip(args.!ashtml)=1 then do                /* UPDATER interactive mode? */
  astext=0
  call lineout tempfile,'<html><head><title>Select a file to retrieve </title></head>'
  call lineout tempfile,'<body 'back1'> '
  call lineout tempfile,'<h3>UpDater: View &amp; retrive files (one at a time) </h3>'
  CALL LINEOUT TEMPFILE,'Select a file to retrieve from the <b>'||strip(translate(args.!package))'</b> package'
  call lineout tempfile,'<em>(on 'servername') </em> '
  call lineout tempfile,'<hr width="30%">'
  call lineout tempfile,'<em>There are <a name="top"> <tt>'package_opts.0'</tt> </a>sections in 'args.!package':</em>'
  call lineout tempfile,'<table><tr>'
  mn=0
  do jj=1 to package_opts.0
     mn=max(mn,length(package_opts.jj.!rdir))
  end /* do */
  mn=mn+3
  ncols=trunc(80/mn)
  do jj=1 to package_opts.0
       call lineout tempfile,'<td><a href="#sec'||jj||'">'package_opts.jj.!rdir'</a>&nbsp;&nbsp;&nbsp;</td>'
       if jj//ncols=0 then call lineout tempfile,'<tr>'
  end /* do */
  call lineout tempfile,'</table><hr width="30%">'
  xtras=''
  if args.!rsync=1 then xtras='&rsync=1'
  if args.!gzip=1 then xtras=xtras||'&gzip=1'
end
else do                         /* HEADER LINE FOR "list of files" */
   aline='Package='args.!package
   aline=aline||' '||'Server='servername
   if package_opts.!encrypt<>"" then aline=aline' Encrypt='||translate(package_opts.!encrypt,',',' ')
   aline=aline||' '||'CREATED='nowtime
   if candogzip=1 then aline=aline' GZIP=1'
   if candorsync=1 then aline=aline' RSYNC=1'
   if package_opts.!Privs<>'' & pos('*',package_opts.!Privs)=0 then 
          aline=aline||' NEEDPRIVS=1'
   aline=aline' SELECTOR='areq
   aline=aline||' '||'PKG_DATE='package_opts.!PKG_DATE
   if args.!client=1 then call add_remote_client_stuff

   if not_send_pieces<>1 then do          /* do not send in pieces */
      do_multi=1
      ah='X-updater: 'xlist
      rcode=sref_multi_send(aline||'0d0a'x,'application/x-updater','1S',0,0,0,ah)
      if rcode<0 then do
         if verbose>1 then call pmprintf('UpDater: connection broken ('rcode')?')
         broken=1
         return 0
      end 
   end
   else do              /* do not send in pieces */
      call lineout tempfile,aline
   end

/* save to "cached" of cachefile */  
   if astext=1 & package_opts.!cache<>0 & selects.!ALL='' then do
         foo=sysfiledelete(package_opts.!cachefile)
         foo=stream(package_opts.!cachefile,'c','open write')
         if abbrev(strip(translate(foo)),'READY')=1 then do  /* cache file open */
            cachefile=package_opts.!cachefile
            call lineout cachefile,aline 
         end 
    end


end /* do */

/* now go to each entry in package, find files, toss exclusions, and write info to
"list of files" */

ndid=0
do jj=1 to package_opts.0
   rdir=strip(package_opts.jj.!rdir)
   if (astext=0) then do
      if jj>0 then
         call lineout tempfile,'<font size="+2"><a name="sec'||jj'">'rdir'</a></font> (<a href="#top">top</a>)'  
      else
         call lineout tempfile,'<font size="+2"><a name="sec'||jj'">'rdir'</a></font> '  
      call lineout tempfile,'<ul>'  
   end
   sss=package_opts.jj.!subs
   dirget=package_opts.jj.!cfile

   foo=translate(dirget,' ','\')
   aa=word(foo,words(foo))
   if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then dirget=dirget'\*.*'

   thisdir=strip(filespec('d',dirget)||filespec('p',dirget),,'\')

/* make sure directory exists */
   if dosisdir(thisdir)=0 then do
       call pmprintf('UpDater warning: 'args.!package', no such directory 'package_opts.jj.!cfile )
   end /* do */

   nthisdir=length(thisdir)
   csss=''; if sss=1 then csss='S'
   a1=sysfiletree(dirget,flist,'FT'||CSSS)
   ndid=ndid+flist.0
   do jj2=1 to flist.0
        parse var flist.jj2 adate asize . aname ;aname=strip(aname)
        if left(adate,2)<70 then 
            adate='20'||adate
        else
            adate='19'||adate

        KK=CHECK_EXCLUDE(ANAME)
        IF KK=1 then ITERATE

        aname2=delstr(aname,1,nthisdir)
        aname3=translate(aname2,'/','\')

/* if "client selection of files" (getl4), see if this file is permitted */
        if selects.!ALL=0 then do               /* not all selected */
           if selects.jj<>1 then do        /* not all in dir selected */
                goo2=translate('!'aname3)
                if selects.jj.goo2<>1 then iterate /* this file is NOT selected, so iterate */
           end /* do */
        end /* do */

/* HERE IS WHERE A CLIENT (non-html) LIST IS CONSTRUCTED */
        if astext=1 then do
           acrc=''
           noce=check_noencode(aname)
           if addcrc=1 then do
               acrc=fileadler(aname)
           end
           lline=adate||' '||jj||' '||aname3||' 'rdir' 'noce' 'acrc' 'asize

           if do_multi=1 then do
              rcode=sref_multi_send(lline||'0d0a'x,,'1A',0)
              if rcode<0 then do
                if verbose>1 then call pmprintf('UpDater: connection broken ('rcode')?')
                 broken=1 
                 return 1
              end 
           end 
           else do
              call lineout tempfile,lline
           end
           if cachefile<>'' then do
              call lineout cachefile,adate||' '||jj||' '||aname3||' 'rdir' 'noce' 'acrc' 'asize
           end

        end

/* html response */
        else do         
           modddate=''
           if args.!nodate<>1 then moddate='&mod='||adate
           aurl='<a target="UpDater" href="/updater?file='||aname3||'&entry='||jj||'&package='||args.!package||'&action=file1' ,
                ||xtras||'">'
           call lineout tempfile,'<li> '||aurl||aname3||'</a> '
        end /* do */

   end /* do */
   if (astext=0) then do
      call lineout tempfile,'</ul>'  
   end
end /* do */
if astext=0 then do
        call lineout tempfile,'<a href="/updater?">Return to updater main menu</a> (at 'servername')'
        call lineout tempfile,'</body></html>'
end /* do */
call lineout tempfile

/* close cache file? */
if cachefile<>'' then foo=stream(cachefile,'c','close')

return ndid




/*************************************/
/*  create a seletion list */

do_getl3:                             

/* find all files matching each  .!cfile entry */

/* send  as an html list (for browser-based human processing */
iserror=0
broken=0
do_multi=0


call lineout tempfile,'<html><head><title>Select  files to retrieve </title></head>'
call lineout tempfile,'<body 'back1'> '
  call lineout tempfile,'<h3>UpDater: Selective file &amp; directory retrieval  </h3>'
  CALL LINEOUT TEMPFILE,'<em>For the <b>'||strip(translate(args.!package))'</b> package'
  call lineout tempfile,'(on 'servername') </em> '

  call lineout tempfile,'<FORM action="/UPDATER" method="POST" name="DO_GETL3">'
  call lineout tempfile,'<INPUT TYPE="hidden" name="ACTION" value="GETL4">'
  call lineout tempfile,'<input type="hidden" name="PACKAGE" value="'||args.!package'">'
  call lineout tempfile,'<input type="hidden" name="CLIENT" value="1">'
  if args.!GETUSER=1 then do
    call lineout tempfile,'<input type="hidden" name="GETUSER" value="1">'
  end

  if candorsync=1 then 
     call lineout tempfile,'<input type="hidden" name="RSYNC" value="1">'

  if candogzip=1 then
     call lineout tempfile,'<input type="hidden" name="GZIP" value="1">'
 
/* start of visible components of the form */
  
  call lineout tempfile,'<hr width="30%">'

  call lineout tempfile,'<table><tr valign="top">'
 
  call lineout tempfile,'<td colspan=2 align="MIDDLE"><a name="top"><input type="Submit" value="Get the selected files from 'args.!package'... "></a><br>'
  packagename=args.!package
if args.!GETUSER=1 then do
  call lineout tempfile,'<tr><td valign="top">Username and password  </td>'
  call lineout tempfile,'<td>Username:<input type="text" name="USER" size=10  > '
  call lineout tempfile,'Password: <input type="password" name="PWD"  size=10> '
  call lineout tempfile,'<br><em>The username and password required for the 'packagename' package may be different'
  call lineout tempfile,' than your 'servername'  "logon" username and password</em>'
end


  call lineout tempfile,'<tr valign="top"><td>Destination directory (on your computer)</td> '
  call lineout tempfile,'<td><input type="text" name="DESTDIR"  size=30> </td>'

  call lineout tempfile,'<tr valign="top"><td>Backup directory (on your computer).</td> '
  call lineout tempfile,'<td><input type="text" name="BACKUPDIR" size=30>'
  call lineout tempfile,'<br><em> To suppress backing up changed files, leave this field empty.</em></td>'

  call lineout tempfile,' <tr valign="top"><td>' ,
     '<input type="checkbox" name="ALL" value="1"> <em>Retrieve all the files</em></td>'

  call lineout tempfile,'<td>'
  call lineout tempfile,'If you retrieve all the files:'
  call lineout tempfile,'the <em>file checkboxes</em> and the'
  call lineout tempfile,'<em>section checkboxes</em> will be ignored.<br>'
  call lineout tempfile,'If you check a <b>section</b>:  '
  call lineout tempfile,'the <em>file checkboxes</em> under that <b>section</b> will be ignored.'


  call lineout tempfile,' </td><tr>'

  call lineout tempfile,'<td><em>Retrieve all the files from several of the <tt>'package_opts.0'</tt> sections in 'args.!package'</em>'
  call lineout tempfile,'</td>'

  call lineout tempfile,'<td>'
  call lineout tempfile,'<table border=1 ><tr>'
  mn=0
  do jj=1 to package_opts.0
     mn=max(mn,length(package_opts.jj.!rdir))
  end /* do */
  mn=mn+3
  ncols=trunc(60/mn)

  ckl=''; multis=''
  do jj=1 to package_opts.0
       nowrdir=package_opts.jj.!rdir
       if wordpos(translate(nowrdir),ckl)>0 then 
          multis=multis' 'nowrdir
       else
         ckl=ckl' 'nowrdir
  end
  do jj=1 to package_opts.0
       call lineout tempfile,'<td>' 
       call lineout tempfile,'<input type="checkbox" name="DIR_'jj'" value="1">'
       nowrdir=package_opts.jj.!rdir
       if wordpos(translate(nowrdir),multis)>0 then do
          call lineout tempfile, '<a href="#sec'||jj||'">'nowrdir'(<font size=-1>'jj'</font>)</a>&nbsp;&nbsp;&nbsp;</td>'
       end
       else do
          call lineout tempfile, '<a href="#sec'||jj||'">'nowrdir'</a>&nbsp;&nbsp;&nbsp;</td>'
       end
       if jj//ncols=0 then call lineout tempfile,'<tr>'
  end /* do */
  

  call lineout tempfile,'</table>'
  call lineout tempfile,'</td>'
  call lineout tempfile,'</table><br>'

  call lineout tempfile,'<em>Select specific files to update .... </em><br>'

/* now start displaying list of directories */
ndid=0
do jj=1 to package_opts.0


   rdir=strip(package_opts.jj.!rdir)

/* write subdir name */
/*   call lineout tempfile,'<input type="checkbox" name="DIR_'jj'" value="1">'*/
   if wordpos(translate(rdir),multis)>0 then do
      call lineout tempfile,'<font size="+2"><a name="sec'||jj'">'rdir'(<font size=-1>'jj'</font>)</a></font> (<a href="#top"><font size=-1>top</font></a>)'  
   end
   else do
      call lineout tempfile,'<font size="+2"><a name="sec'||jj'">'rdir'</a></font> (<a href="#top"><font size=-1>top</font></a>)'  
   end

/* prepare to writes files in this subdir */
   call lineout tempfile,'<ul>'  
   sss=package_opts.jj.!subs
   dirget=package_opts.jj.!cfile

   foo=translate(dirget,' ','\')
   aa=word(foo,words(foo))
   if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then dirget=dirget'\*.*'

   thisdir=strip(filespec('d',dirget)||filespec('p',dirget),,'\')

/* make sure directory exists */
   if dosisdir(thisdir)=0 then do
       call pmprintf('UpDater warning: 'args.!package', no such directory 'package_opts.jj.!cfile )
   end /* do */

   nthisdir=length(thisdir)
   csss=''; if sss=1 then csss='S'
   a1=sysfiletree(dirget,flist,'FT'||CSSS)
   ndid=ndid+flist.0

   do jj2=1 to flist.0          /* start writing files in this subdir */
        parse var flist.jj2 adate  asize . aname ;aname=strip(aname)
        if left(adate,2)<70 then 
            adate='20'||adate
        else
            adate='19'||adate

       csize=fixup_size(asize,use_abbrev)

        parse var adate aday +10 . +1 atime
        KK=CHECK_EXCLUDE(ANAME)
        IF KK=1 then ITERATE
        aname2=delstr(aname,1,nthisdir)
        aname3=translate(aname2,'/','\')
        
        modddate=''
        aurl='<input type="checkbox" name="F_'jj'_'jj2'" value="'||aname3||'">'
        call lineout tempfile,'<li> '||aurl||'  '||aname3||'&nbsp;&nbsp;<tt>'csize'  'aday' '||translate(atime,':','/')||'</tt>'

   end /* do */
   call lineout tempfile,'</ul>'  
end /* do */
  call lineout tempfile,'<br><input type="Submit" value="Get the selected files from 'args.!package'... ">'
call lineout tempfile,'<input type="reset" value="Reset the form">'
call lineout tempfile,'</form>'

call lineout tempfile,'<a href="/updater?">Return to updater main menu</a> (at 'servername')'
call lineout tempfile,'</body></html>'
call lineout tempfile

return ndid



/********/
/* pull stuff from cache header line, save in hdr.! */
get_CREATED:procedure expose hdr.
parse arg aline0
created=0 ; pkg_date=0
DO UNTIL ALINE0=''
  parse var aline0 a1 aline0
  parse var a1 avar '=' aval; avar=strip(translate(avar))
  if avar='CREATED' then created=strip(aval)
  if avar='PKG_DATE' then pkg_date=strip(aval)
end
return created' 'pkg_date
return 0

/************/
/* add remote client stuff to "aline" */
add_remote_client_stuff:
      if auser<>'' then do
         aa=auser':'apwd
         aline=aline' USERPWD='aa
      end /* do */
      if ddir<>'' then do
         aline=aline' DESTDIR='||ddir
      end
      if fixit(args.!BACKUPDIR)<>'' then do
         aline=aline' BACKUPDIR='||fixit(args.!BACKUPDIR)
      end
      if args.!DODELETE<>'' then do
         aline=aline' DODELETE='args.!DODELETE
      end
      if args.!CHECK<>'' then do
         aline=aline' CHECK='args.!CHECK
      end
      rng1=strip(fixit(args.!range1))
      if rng1<>'' then do
         rng2=strip(fixit(args.!range2))
         if datatype(rng1)='NUM' & datatype(rng2)='NUM' then do
             rng12=strip(rng1)||':'||strip(rng2)
             aline=aline' RANGE='||rng12
         end /* do */
      end
      if args.!QUIET=1 then aline=aline' QUIET=1 '
      if args.!dORUN=1 then  do
           if package_opts.!RUN<>'' then do
               parse var package_opts.!run aprog atype runargs
               ctype=translate(strip(atype))
               istype=''
               if ctype='PM' then do
                  istype='PM'
               end /* do */
               else do
                  parse var package_opts.!run . runargs
               end /* do */
               isargs=''
               if abbrev(translate(strip(runargs)),'ARG=')=1 then do
                  parse var runargs . '=' isargs
               end /* do */
               aline=aline' RUN='||strip(aprog)
               if istype<>'' then aline=aline' RUNTYPE='||istype
               if isargs<>'' then aline=aline' RUNARGS='||mkpack64(isargs)
           end /* do */
       end

      return 1
   

/**********************/
mkPACK64:procedure
parse arg mess

biga=xrange('A','Z')||xrange('a','z')||xrange('0','9')||'+/'

s2=x2b(c2x(mess))

nith=trunc((length(s2)/6)+.9)
cont=copies(' ',nith)
oof=""
do mm=0 to 63
      oof=oof||x2c(b2x(right('00'||x2b(d2x(mm)),6)))
end /* do */
do ith=1 to nith 
  a1=substr(s2,(ith*6)-5,6,0)
  cont=overlay(x2c(b2x(a1)),cont,ith) 
end /* do */
pint=""
pint=translate(cont,biga,oof)
j1=length(pint)//4
if j1<>0 then pint=pint||copies('=',4-j1)
return pint

/************/
/* clean up an argument */
fixit:procedure
parse arg aa
aa=translate(aa,' ','+'||'0d0a0900'x)
aa=packur(aa)
aa=strip(aa)
return aa


/*************/
/* CHECK A FILEname against an exclusion list */
check_exclude:procedure expose package_opts.
parse arg afile
foo=0
do mm=1 to package_opts.!exclude.0
   ex1=package_opts.!exclude.mm
   sub1=package_opts.!exclude.mm.1
 
   if (sub1<>1) then do         
      adir=translate(filespec('d',afile)||filespec('p',afile))
      xdir=translate(filespec('d',ex1)||filespec('p',ex1))
      if adir<>xdir then iterate   /* skip if not same directory */
      aname=filespec('n',afile)
      xname=filespec('n',ex1)
      foo=sref_wild_match(aname,xname)
    end
    else do             /* subdir matches okay */
      foo=sref_wild_match(afile,ex1)
    end                 /* sub1 */
    if foo<>0 then return 1           /* match found, so exclude */
end /* do */
return 0                /* if here, no exclusions */

/*************/
/* CHECK A FILEname against an  noencode list */
check_noencode:procedure expose package_opts.
parse arg afile

foo=0
do mm=1 to package_opts.!noencode.0
   ex1=package_opts.!noencode.mm
   sub1=package_opts.!noencode.mm.1
 
   if (sub1<>1) then do         
      adir=translate(filespec('d',afile)||filespec('p',afile))
      xdir=translate(filespec('d',ex1)||filespec('p',ex1))
      if adir<>xdir then iterate   /* skip if not same directory */
      aname=filespec('n',afile)
      xname=filespec('n',ex1)
      foo=sref_wild_match(aname,xname)
    end
    else do             /* subdir matches okay */
      foo=sref_wild_match(afile,ex1)
    end                 /* sub1 */
    if foo<>0 then return 1           /* match found, so do not encode */
end /* do */
return 0                /* if here, no noencodes */



/*************/
/* construct a list of packages.
   First line: # of packages
   2.,#Packages+1 lines: package name descripton (if requested)
*/


do_packagelist:

astext=1
if strip(args.!ashtml)=1 then astext=0
if astext=0 then do
  call lineout tempfile,'<html><head><title>Currently available UpDater packages </title></head>'
  call lineout tempfile,'<body 'back1'>'
  call lineout tempfile,'<h3>Welcome to UpDater (beta 1.10)</h3>'
end

 foo=sysfiletree(package_dir'\*.pkg',plist,'FT')
 if astext=1 then do
   call lineout tempfile,plist.0
 end
 else do
    call lineout tempfile,'<blockquote>This UpDater <em>central server </em> &nbsp; (at <tt> 'servername')</tt> '
    call lineout tempfile,'currently supports <tt>'plist.0'</tt> <em>packages</em></blockquote><table>'
 end

 do mm=1 to plist.0
       adesc=''
       parse var plist.mm adate . . nname ;nname=strip(nname)
       if left(adate,2)<70 then 
            adate='20'||adate
       else
            adate='19'||adate
       nname0=filespec('n',nname)
       if strip(translate(args.!package))='!LIST' then do  /* extract description */
            isize=stream(nname,'c','query size')
            if (isize=0 | isize='') then do    /* bad package */
                if astext=1 then
                   call lineout tempfile,'; unavailable package: 'nname0
                else
                   call lineout tempfile,'<tr><td>'mm'</td><td><tt>'nname0'</tt></td><td><em> unavailable </em> </td>'

                iterate
            end /* do */
            stat=read_package(nname,isize)             /* read in package file */
            foo=stream(nname,'c','close')
            adesc=package_opts.!title
       end /* do */
       nname0a=nname0
       foo=lastpos('.',nname0)
       if foo>1 then nname0a=left(nname0,foo-1)
       if astext=1 then do
          call lineout tempfile,adate' 'nname0a' 'adesc
       end
       else do
          call lineout tempfile,'<tr><td>'mm'</td><td><a href="/updater?package='nname0a'&action=display">'nname0a'</a> </td><td> 'adesc '</td>'
       end
 end /* do */
 if astext=0 then do
    call lineout tempfile,'</table>'
    call lineout tempfile,'<hr>Would you like to read the <a target="info" href="/updater?send=UPDATER.DOC&text=1">UpDater documentation</a>'
    call lineout tempfile,'</body></html>'
 end
 call lineout tempfile
 return 1


/*************************************/
/* formatted display of  the various entries in this package */
disp_package:
  call lineout tempfile,'<html><head><title>Displaying an UpDater package </title></head>'
  call lineout tempfile,'<body 'back1'>'
  CALL LINEOUT TEMPFILE,'<h3>Contents of Updater Package: '||strip(translate(args.!package))'</h3>'
  call lineout tempfile,'<ul><li><b>Title</b>: 'package_opts.!title
  call lineout tempfile,'<li><b>Description</b>: 'package_opts.!desc
  if package_opts.!Privs<>'' then
     call lineout tempfile,'<li><b>Privileges</b>: 'package_opts.!privs
  call lineout tempfile,'</ul>'
  call lineout tempfile,'<h3>You can ... </h3>'

  call lineout tempfile,'<menu>'
  call lineout tempfile,'<li>'
  if package_opts.!privs='' | pos('*',package_opts.!privs)>0 then
    call lineout tempfile,'<a href="/updater?package='||args.!package||'&action=getl2&client=1">'
  else
    call lineout tempfile,'<a href="/updater?package='||args.!package||'&action=getl2&client=1&GETUSER=1">'
  call lineout tempfile,'UpDate these files</a> to your computer '
  call lineout tempfile,'(using the <a target="info" href="/updater?action=UPCLIENTDOC">UpClient helper app</a>) '
  call lineout tempfile,'<li> or ... '
  if package_opts.!privs='' | pos('*',package_opts.!privs)>0 then
    call lineout tempfile,'<a href="/updater?package='||args.!package||'&action=getl3&client=1">'
  else
    call lineout tempfile,'<a href="/updater?package='||args.!package||'&action=getl3&client=1&GETUSER=1">'
  call lineout tempfile,'selectively UpDate these files</a> to your computer '

  call lineout tempfile,'<li>One-by-one view  &amp; select the '
  call lineout tempfile,'<a href="/updater?package='||args.!package||'&action=getl&ashtml=1">files',
                       ' in 'args.!package '</a>?'
  call lineout tempfile,'<li><a href="/Updater?">Return to UpDater main menu</a> (at 'servername')'

  call lineout tempfile,'</menu>'
  call lineout tempfile,'<h3>Description of the 'args.!package' package</h3>'
  call lineout tempfile,'<table><tr><th>Remote Dir </th><th>Central Files </th><th>Subdirectories? </th>'
  do jj= 1 to package_opts.0
     sok='' 
     if package_opts.jj.!subs=1 then sok='YES'
     a1=package_opts.jj.!rdir
     a2=package_opts.jj.!cfile
     call lineout tempfile,'<tr><td> 'a1'</td><td> 'a2'</td><td> 'sok'</td></tr>'
  end /* do */
  call lineout tempfile,'</table>'
  if package_opts.!exclude.0>0  then do
    call lineout tempfile,'<hr width="30%">'
    call lineout tempfile,'<b>File exclusion rules</b><br>'
    call lineout tempfile,'<table><tr><th>File Rule</th><th>Subdirectories? </th>'
    do jj= 1 to package_opts.!exclude.0
       sok='' 
       if package_opts.!exclude.jj.1=1 then sok='YES'
       a2=package_opts.!exclude.jj
       call lineout tempfile,'<tr><td> 'a2'</td><td> 'sok'</td></tr>'
    end 
  end
  call lineout tempfile,'</table>'
  call lineout tempfile,'</body></html>'
  call lineout tempfile
  return 1

  end

/***********/
/* describe the upclient helper app */
tell_upclient:
  call lineout tempfile,'<html><head><title>The UpClient helper app </title></head>'
  call lineout tempfile,'<body 'back1'>'
  call lineout tempfile,'<a name="upclient"><h3>The UpClient helper app </h3></a>'
  call lineout tempfile,'The <a href="/updater?send=UPCLIENT.CMD">UpClient</a> helper application is a REXX program (UPCLIENT.CMD)'
  call lineout tempfile,'that will update a package of files to your computer. '
  call lineout tempfile,'<b>To use UPCLIENT.CMD, you must be running NetScape under OS/2.</b>'
  call lineout tempfile,'<p>Assuming this describes you, after '
  call lineout tempfile,'<a href="/updater?send=UPCLIENT.CMD">downloading UpClient</a>'
  call lineout tempfile,'you should <ol>'
  call lineout tempfile,'<li>Copy UPCLIENT.CMD to a directory in your PATH (i.e.; to x:\OS2\APPS, where x: is your boot drive)'
  call lineout tempfile,'<li>Tell NetScape to use UPCLIENT.CMD with all <tt>application/x-updater</tt> responses'
  call lineout tempfile,'<br>To do this, you should: '
  call lineout tempfile,'<menu><li>Open up the Preferences menu in Netscape (under NS 4.6,<tt> Edit-Preferences</tt>)'
  call lineout tempfile,'<li>Select <tt>Applications</tt>'
  call lineout tempfile,'<li>Select the <tt>New Type</tt> option'
  call lineout tempfile,'<li>Fill in the fields of the <tt>New Type</tt> screen with: '
  call lineout tempfile,'<ul><li><u>Description of type</u>: UpDater Client '
  call lineout tempfile,'<li><u>File Extension</u>: UPD '
  call lineout tempfile,'<li><u>MIME Type</u>: application/x-updater'
  call lineout tempfile,'<li><u>Application to use</u>: cmd.exe /c "x:\os2\apps\upclient.cmd"'
  call lineout tempfile,'<br> <em>be sure to include the " characters </em>'
  call lineout tempfile,'</ul>'
  call lineout tempfile,'</menu>'
  call lineout tempfile,'</ol>'
  call lineout tempfile,'Once UpClient is installed, when you UpDate a package from a'
  call lineout tempfile,'<Em>UpDater central server</em>, Netscape will run UpClient in '
  call lineout tempfile,'seperate window.  After asking a few questions, UpClient will'
  call lineout tempfile,'GET files and write them to your hard drive.'
  call lineout tempfile,'<p>Alternatively, <a href="/updater?send=updateme.cmd">UpDateMe.cmd</a>, '
  call lineout tempfile,'a non-gui stand-alone program, can be used as your client-side interface'
  call lineout tempfile,'to a central server.'

  call lineout tempfile,'<hr width="30%">'
  call lineout tempfile,'Note: UpClient works best when you have the following libraries installed:'
  call lineout tempfile,'<ol>'
  call lineout tempfile,'<li> <a href="/updater?send=rxRsync.dll">The rxRsync library</a>'
  call lineout tempfile,'<li> <a href="/updater?send=rxgzlib.dll">The rxgzlib library</a>'
  call lineout tempfile,'<p><em>(install these .dll files in your <TT>LIBPATH</tt>. For example, in x:\OS2\DLL)</em>'

  call lineout tempfile,'</ol>'

  call lineout tempfile,'</body></html>'

  call lineout tempfile
  return 1

/*************************************/
/* check so see if client (remote server) has rights to this package */

check_privs:procedure expose tempfile args. servername
parse upper arg privset,daprivs
IF WORDPOS('SUPERUSER',PRIVSET)>0 then RETURN 'OK' 
do mm=1 to words(daprivs)
   ap=strip(word(daprivs,mm))
   if wordpos(ap,privset)>0 then return 'OK'
end /* do */

  'header add WWW-Authenticate: Basic Realm=<Manager>'  /* challenge */
   call lineout tempfile, '<!doctype html public "-//IETF//DTD HTML 3.0//EN">'
   call lineout tempfile,'<html><head><title>UpDater. Please enter a valid username</title></head> '
   call lineout tempfile,' <strong>'
   call lineout tempfile,' To acquire 'args.!package', a valid  username is required</strong>'
   call lineout tempfile,' </body> </html> '
   call lineout tempfile
   return sref_response('unauth'," UpDater-- requires a valid username/password ",servername,1)

/*************************************/
/* read package file into package_opts. stem */
read_package:procedure expose package_opts. crlf candorsync candogzip
parse arg afile,isize

afilea=charin(afile,1,isize)||'0d0a'x||' '
foo=stream(afile,'c','close')

doing=''
package_opts.=''
package_opts.!exclude.0=0
package_opts.!noencode.0=0
package_opts.!cache=0

parse var afile afil0 '.' .
afilc=afil0'.PKC'

if stream(afilc,'c','query size')=0 then foo=sysfiledelete(afilc)  /* 0 length files are trouble */

package_opts.!PKG_DATE=make_timestamp(afile)
package_opts.!cachefile=afilc
package_opts.!cachestamp=make_timestamp(afilc)  /* timestamp of "cache" entry */

package_opts.0=0

do until afilea=''
  parse var afilea aline (crlf) afilea
  aline=strip(aline)

  select 
    when abbrev(aline,';')=1 then iterate   /* skip comments */
    when aline='' & doing='' then iterate   /* skip multiple blank lines */
    when aline='' then do           /* end of a entry */ 
       doing=''                   /* get ready for next entry */
       iterate
    end
    when doing='' then do           /* find name: */
      if pos(':',aline)=0 then iterate  /* error, ignore */
      parse var aline aname ':' aline
      doing=strip(translate(aname))
      
    end
    OTHERWISE NOP
   END
   if aline='' then iterate
   SELECT
    when abbrev(doing,'PRIV')=1 then do   /* not a FILES entry */
        package_opts.!PRIVS=package_opts.!PRIVS' 'aline
        iterate
    end
    when abbrev(doing,'DESC')=1 then do   /* not a FILES entry */
        package_opts.!DESC=package_opts.!DESC' 'aline
        iterate
    end
    when abbrev(doing,'TITLE')=1 then do   /* not a FILES entry */
        package_opts.!TITLE=package_opts.!TITLE' 'aline
        iterate
    end

    when abbrev(doing,'FILE')+abbrev(doing,'INC')>0 then do   /* an include entry */
       parse var aline remote_dir central_file alsosubs .
       inn=package_opts.0+1
       package_opts.inn.!rdir='/'||strip(translate(strip(translate(remote_dir)),'/','\'),,'/')

       ifile=translate(strip(translate(central_file)),'\','/')

       foo=translate(ifile,' ','\')
       aa=word(foo,words(foo))
       if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then ifile=ifile'\*.*'

       package_opts.inn.!cfile=ifile
       if translate(strip(alsosubs))='/S' then package_opts.inn.!subs=1
       package_opts.0=inn
    end

    when abbrev(doing,'EXC')=1 then do   /* a EXCLUDE entry */
       parse var aline  exc_file alsosubs .
       inn=package_opts.!exclude.0+1
       efile=translate(strip(translate(exc_file)),'\','/')

       foo=translate(efile,' ','\')
       aa=word(foo,words(foo))
       if (pos('.',aa)+pos('*',aa)+pos('?',aa))=0 then efile=efile'\*.*'

       package_opts.!exclude.inn=efile
       if translate(strip(alsosubs))='/S' then package_opts.!exclude.inn.1=1
       package_opts.!exclude.0=inn
    end

    when abbrev(doing,'NO_E')=1 then do   /* a no_encode entry */
       parse var aline  noe_file alsosubs  .
       inn=package_opts.!noencode.0+1
       package_opts.!noencode.inn=translate(strip(translate(noe_file)),'\','/')
       if translate(strip(alsosubs))='/S' then package_opts.!noencode.inn.1=1
       package_opts.!noencode.0=inn
    end

    when abbrev(doing,'ENCRYPT')=1 then do
        package_opts.!encrypt=translate(strip(space(aline)))
    end 

    when abbrev(doing,'RUN')=1 then do
        package_opts.!run=strip(aline)
    end 

    when abbrev(doing,'CACHE')=1 then do   /* a no_encode entry */
       parse var aline  adays . ; adays=strip(adays)
       if datatype(adays)='NUM' then  package_opts.!cache=adays
    end

    otherwise do
      AOPT='!'||DOING
      package_opts.aopt=package_opts.aopt||' '||aline
      iterate
    end /* do */

  end         /*select */
end             /* do loop */
return 1


/****/
/* make a nnn.xxx timestamp */
make_timestamp:procedure
parse arg afile
foo=sysfiletree(afile,'ttt','FT')
if ttt.0=0 then return 0                /* no time stamp */
parse var ttt.1 yr '/' mon '/' day '/' hr '/' mnutes .
if yr<70 then 
  yr=2000+yr
else
  yr=1900+yr
tt2=dateconv(yr||mon||day,'S','B')
numeric digits 10
tt3=((hr*60)+mnutes)/1440
tstamp=tt2+tt3
return tstamp 

/***********/
/* adler-32 crc of file aname */
fileadler:procedure expose candorsync
parse arg aname

if candorsync=0 then return ''

foo=stream(aname,'c','open read')
if abbrev(strip(translate(foo)),'READY')=0 then return ''

sz=stream(aname,'c','query size')
if sz='' | sz=0 then return ' '
aa=charin(aname,1,min(100000,sz))
adler=rx_adler32(aa)
igot=100000
do forever
   if igot>=sz then leave
   iget=min(100000,sz-igot)
   aa=charin(aname,,iget)
    adler=rx_adler32(aa,adler)
    igot=igot+100000
end
foo=stream(aname,'c','close')

return adler



/*** jump here on error */
badguys:
say "updater error "Rc " at " sigl
if completed()=0 then do
    'header add x-updater: error'
    'string error 'rc ' at ' sigl
end

return ' '


/********************/
/* check for gzip and rsync dlls -- note that we assume rxsock, rexxutil,
   and rexxlib are available */

check_dlls:

if rxfuncquery('rx_md4')=1  then do
  foo=RXFuncAdd( 'RXRsyncLoad', 'RXRSYNC', 'RxRsyncLoad')
  if foo=0 then call RxRsyncLoad
end
if rxfuncquery('rx_md4')=1  then do
   candorsync=0
   call pmprintf("UpDater warning: unable to find rxrsync")
end

if rxfuncquery('rxgzLoadFuncs')=1  then do
  foo=RXFuncAdd( 'rxgzLoadFuncs', 'rxgzLIB', 'rxgzLoadFuncs')
  if foo=0 then call rxgzLoadFuncs
end
if  rxfuncquery('rxgzDeflateFile')=1  then do
   candogzip=0
   call pmprintf("UpDater warning: unable to find rxgzlib")
end

return 0


/* ----------------------------------------------------------------------- */
/* REPLACESTRG:
 newstring=replacestrg(haystack,needle,replacement,type,exact)
  Arguments:
                astring : the "haystack" to look in
                target: the "needle" to look for
                putme: the "new needle" to replace the "needle" with
                type : The direction/type of search
                        FORWARD, BACKWARD, ALL
                exact: YES-- then cases in needle and haystack must match

      Note taht regardless of value of exact, cases are retained in both
      astring and putme.

   Returns the modified astring, or the unmodified astring if target could
   not be found.
*/
/* ----------------------------------------------------------------------- */

replacestrg:

exactmatch=0
backward=0 ; doall=0

parse arg astring ,  target   , putme , type , exactmatch

type = translate(type)
if type="BACKWARD" then backward="YES"
if type="ALL" then doall="YES"

iat=1
joelen=length(target)
joelen2=length(putme)

doagain:                /* here if doall=yes */
 if exactmatch="YES" then do
    if   backward="YES" then
        joe= lastpos(target,astring)
    else
        joe= pos(target,astring,iat)
 end
 else do
   if   backward="YES" then
        joe= lastpos(translate(target),translate(astring))
    else
        joe= pos(translate(target),translate(astring),iat)
 end
 if joe=0 then
         return astring

 astring=delstr(astring,joe,joelen)
 if putme<>' ' then
    astring=insert(putme,astring,joe-1)

 if doall="YES" then do
     iat=joe+joelen2
     signal doagain
 end
/* else, all done */
 return astring

/* fix up size, given format */
fixup_size:procedure
parse arg asize,isabbrev
if isabbrev=1 then do
   if asize>=1000000 then
        return format(asize/1000000,,1)||'M'
   if asize>=1000 then 
      return format(asize/1000,,0)||'K'
end
/* convert to xxx,yyy,zzz */
 il=length(asize)
 if il>3 then do
           oop=""
           do mm=il to 3 by -3
               tt=substr(asize,mm-2,3)
               if mm=il then
                  oop=tt
               else
                 oop=tt||','||oop
           end /* do */
           if mm<>0 then oop=substr(asize,1,mm)||','||oop
           asize=oop
 end
 return asize    /* not abbrev, or < 1000 */


