/* 2 Jan 2000. Daniel Hellerstein (danielh@crosslink.net)
  The Updater Daemon -- obtains files from an updater
  "central server", and saves to a local hard drive.

 This daemon is typically invoked by RUPDATER.CMD (when RUPDATER
 is being run as an SRE-http addon). However, ambitious users 
 can run this directly from the command line, using the following
 syntax:
 
  D:\UPDATER>UpDaemon listfile userpwd destdir tcheck dodel asem backdir range key

where:

   listfile: a fully qualified file name, pointing to a "list of files in a 
            package", as may be returned UPDATER.CMD.
   userpwd: A username and password, seperated by a :
   destdir: The "root" directory -- files obtained from the central server will
            be copied to this directory, or to subdirectories under this
            root directory.
   tcheck: Time/content check. 
             If 0, get all files. 
             If 1, check the datestamp and get if remote version is older (or
             if no remote version)
             If 2, check the crc-32 and get if remote version is different (or
             if no remote version)
   dodel: If 1, then delete files in destdir that are NOT in the listfile.
   asem: A semaphore used to communicate with the daemon. For running from the
         command line, we recommend setting this to 0
   backdir: A "backup" directory.  Files that are being replaces will be 
           "backed up" to this directory. Enter a 0 to suppress this 
           (i.e.; to NOT backup files).
   range:  A range of the "listfile" entries to use. If 0, use all entries.
           Otherwise, range should have the form   i1:i2, where i1 and i2
           are integers. For example:   120:300
      key: A "shared-secret" decryption key.

Uses "package listing" files, as supplied in the command line.

*/


/**** Begin User changeable parameters ****/

/* Archive old "status" reports in  REPORTS subdirectories (of the SITE\site_nickname
   subdirectory.  1 to archive, 0 to not archive */
archive_reports=0

/* allow use of rsync encoding */
candorsync=1

/* allow use of GZIP encoding */
candogzip=1

/* set of "encryption" specs. ENCRYPT.!BLOWFISH is an example,
  you can delete it if you will never use blowfish */
ENCRYPT.!blowfish='bf -q -d %in %out "%key"'

/* number of seconds to wait between socket calls. Large values
   may be  needed over very slow lines (this is NOT the total
   time required to get a file, it's the max seconds for each
   1000 byte "sockrecv" call) */
timeout=30

/* verbosity (using PMPRINTF): 0=none, 1=some, 2= more */
verbose=2


/**** End of User changeable parameters ****/

signal on error name badguys; signal on syntax name badguys
crlf = '0d0a'x

foo=time('r')

parse arg  stuff
parse var stuff aopt listfile user destdir timecheck dodelete asem backupdir arange key .
parse var user auser ':' apwd

if aopt<>'OBTAIN' then do
   say " This UpDater daemon is not meant to be run in stand-alone mode. "
   exit
end  /* Do */

temp_dir=value('TEMP',,'os2environment')


norexxlib=0             /* assume rexxlib is loaded */
call check_dlls         /* load some dlls */

/* open output file */
oo=lastpos('.',listfile)
if oo=0 then do
   outfile=listfile'.'
end
else do
   outfile=left(listfile,oo)
end /* do */

if archive_reports=1 then do
  if stream(outfile'RPT','c','query exists')>0 then do
     rptdir=filespec('d',outfile)||filespec('p',outfile)||'REPORTS'
     if dosisdir(rptdir)=0 then foo=sysmkdir(rptdir)
     bakfile=rptdir||'\'||filespec('n',outfile)
     tfile=dostempname(bakfile'???')
     if tfile<>'' then foo=dosrename(outfile'RPT',tfile)
     foo=sysfiledelete(outfile'RPT')
  end
end
else do
   foo=sysfiledelete(outfile'RPT')
end 
outfile=outfile'RPT'
foo=stream(outfile,'c','open write')
if abbrev(translate(foo),'READY')=0 then do
   call pmprintf('UpDater Daemon ERROR ('foo'): unable to open status file 'outfile)
   exit
end /* do */

call lineout outfile,'UpDater Status Report on 'date('n')||' '||time('n')
call lineout outfile

/* read the package list */
foo=strip(translate(stream(listfile,'c','open read')))
if abbrev(foo,'READY')=0 then do
   call pmprintf('UpDater Daemon ERROR: can not open 'listfile)
   exit
end 
isz=stream(listfile,'c','query size')
if (isz=0 | isz='') then do
   call pmprintf('UpDater Daemon ERROR: empty or missing 'listfile)
   exit
end 
stuff=charin(listfile,1,isz)
foo=stream(listfile,'c','close')

/* check status of destdir */
if dosisdir(destdir)=0 then do
   foo=sysmkdir2(destdir)
   if foo<>0 then do
       call pmprintf('UpDater Daemon ERROR: unable to create destdir: 'destdir
       exit
   end /* do */
end /* do */

/* check status of backupdir */
if backupdir<>0 & backupdir<>'' then do
  if dosisdir(backupdir)=0 then do
   foo=sysmkdir2(backupdir)
   if foo<>0 then do
       call pmprintf('UpDater Daemon ERROR: unable to create backupdir: 'backupdir)
       exit
   end /* do */
 end /* do */
end
else do
   backupdir=''
end /* do */

parse var stuff hdrline (crlf) stuff
opts.=''
do until hdrline=''
   parse var hdrline aword hdrline
   aword=strip(aword)
   parse var aword atype '=' avalue ; atype=strip(translate(atype))
   if wordpos(atype,'PACKAGE SERVER ENCRYPT GZIP RSYNC')=0 then iterate
   atail='!'||atype
   opts.atail=strip(avalue)
end 

/* check for required fields */
if opts.!SERVER='' then do
       call pmprintf('UpDater Daemon ERROR: no central server info in package list 'listfile)
       exit
end /* do */

if opts.!PACKAGE='' then do
       call pmprintf('UpDater Daemon ERROR: no package name in package list 'listfile)
       exit
end /* do */

/* look for supported encryption */
encrypts=''
key=strip(translate(key))
if opts.!encrypt<>'' & key<>0 then do                 /* see which ones are supported */
   aencrypts=translate(translate(opts.!encrypt,' ',','))
   jj=0
   do mm=1 to words(aencrypts)
        aword=strip(word(aencrypts,mm))
        if symbol('ENCRYPT.!'||aword)='VAR' then do
           encrypts=encrypts','aword
        end /* do */
    end        
    encrypts=strip(encrypts,,',')
end

call lineout outfile,'Package: 'opts.!package 
call lineout outfile,'which is on the server at: ' opts.!server

call lineout outfile,'Package list: 'listfile
call lineout outfile,'Destination Directory: 'destdir
if backupdir<>'' then call lineout outfile,'Backup Directory: 'backupdir

/* open semapore */
if asem<>0 then do
  foo=eventsem_create(asem,'P')
  suicide=translate(asem||'_DIE')
  foo=eventsem_close(suicide)
  foo=eventsem_create(suicide)
end

call syssleep(.2)

/* make a list of files to GET */
mm=0
do until stuff=''
    parse var stuff aline (crlf) stuff
    aline=strip(aline) 
    if abbrev(aline,';')=1 then iterate
    if aline='' then iterate
    mm=mm+1
    parse var aline entries.mm.!date entries.mm.!entry entries.mm.!file entries.mm.!rdir entries.mm.!noce entries.mm.!crc entries.mm!asize .
end /* do */

if mm=0 then do
   call lineout outfile,' '
   call lineout outfile,"ERROR: there are no files specified. " 
   call lineout outfile,"Perhaps there is a problem with the specification of this package? "
   call lineout outfile
   if asem<>0 then do
     foo=eventsem_close(asem)
     foo=eventsem_close(suicide)
   end
   exit
end /* do */

entries.0=mm

servername=get_hostname()   /* address of this machine */

/* now get the files ! */
stat.=0
didit.=0

retrys.=0
retrying=0

len1=length(entries.0)+1


/* verify key */
if key<>0 & key<>'' & encrypts<>'' then do   /* verify that the client's shared secret is correct */ 
    nonce=space(translate(date('n')||time('n')||servername),0)
    tocheck=translate(nonce||key)
    amd4=rx_md4(tocheck)
    req='UPDATER?VERIFY='||nonce';'amd4
    stuff=get_a_url(opts.!server,req,auser,apwd,,1)
    stuff=translate(strip(stuff))
    if abbrev(STUFF,'VERIFY=1')=0 then do
       call lineout outfile,' ERROR: no such decryption key for user 'auser' at 'opts.!server
       call lineout outfile
       if asem<>0 then do
         foo=eventsem_close(asem)
         foo=eventsem_close(suicide)
       end
       exit
    end /* do */
end

i1=1
i2=entries.0
if pos(':',arange)>0 then do
    parse  var arange j1 ':' j2
    if datatype(j1)='NUM' & datatype(j2)='NUM' then do
         i1=j1 ; i2=j2
    end /* do */
end /* do */

tocheck=1+i2-i1
call lineout outfile,'# of files to check: 'tocheck
call lineout outfile,' '

doretry:                /* jump here to retry unGETable files */

didprocess=0

/* open up a socket (and try to re use it as much as possible */
gosock=sockhttp_opensock(,opts.!server)
if abbrev(gosock,'-1')=1 then do
  call pmprintf("Error, unable to create socket to " server)
  exit
end

/* temp file for use by sockhttp */
tempfile=make_tempname('UC1',norexxlib,temp_dir)


do mm0a=i1 to i2
  if retrying=0 then
     mm0=mm0a
  else
     mm0=retrys.mm0a

  if asem<>0 then do
    aa=eventsem_query(suicide)
    if aa=-187 then foo2=eventsem_create(suicide)
    if aa>0 then do
       foo2=eventsem_close(suicide)
       call lineout outfile,'UpDater DAEMON has been killed.'
       call lineout outfile
       if datatype(gosock)='NUM' then   foo=a_sockclose(gosock,1)
       exit
     end 
   end          /* asem<>0 */
   cmm0=right(mm0,len1)': '
   afile=entries.mm0.!file
   afilel=strip(translate(afile,'\','/'),,'\')
   rdir0=strip(strip(translate(entries.mm0.!rdir,'\','/')),,'\')
   if rdir0='' then
      tofile=destdir||'\'||afilel
   else
      tofile=destdir||'\'||rdir0||'\'||afilel
   
   todir=filespec('d',tofile)||filespec('p',tofile)
   if dosisdir(strip(todir,,'\'))=0 then do
       foo=sysmkdir2(todir)
       if foo<>0 then do
          stat.!nocreate=stat.!nocreate+1
          call lineout outfile,cmm0'ERROR could not make directory for 'tofile
          iterate
       end
   end

   backfile=''
   if backupdir<>'' then do
      backfile=backupdir'\'afilel
      backdir=filespec('d',backfile)||filespec('p',backfile)
      if dosisdir(strip(backdir,,'\'))=0 then do
        foo=sysmkdir2(backdir)
        if foo<>0 then do
          stat.!nocreate=stat.!nocreate+1
          call lineout outfile,cmm0'ERROR could not make backup directory for 'backfile
          iterate
       end
     end                /* backupdir exists*/
   end          /* backupdir specified */

   foo='!'||strip(translate(tofile))
   didit.foo=1

/* check date first ? */
   arf=sysfiletree(tofile,'tt','FT')
   oldfile=''; timecheck=strip(timecheck)
   if tt.0<>0 then do           /* compare dates */
        oldfile=tofile
        if timecheck=1 then do                  /* update newer files only */
           parse var tt.1 todate . . .
           if left(todate,2)<70 then
              todate='20'||todate
           else
              todate='19'||todate
           if todate>= entries.mm0.!date then do
                stat.!nochange=stat.!nochange+1
                if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
                call lineout outfile,cmm0'No change (timestamp check) in 'tofile
                iterate
           end /* do */
        end                      /*timecheck */
        if timecheck=2  & candoadler=1 & entries.mm0.!crc<>'' then do          /* compare crc32?*/       
           thiscrc=fileadler(oldfile)
           if strip(translate(thiscrc))=strip(traNslate(entries.mm0.!crc)) then do
                stat.!nochange=stat.!nochange+1
                if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
                call lineout outfile,cmm0'No change (crc32 check) in 'tofile
                iterate
           end /* do */
        end /* do */
   end                /* found old version */

   aentry=entries.mm0.!entry
   req='/Updater?package='opts.!package'&action=file&file='afile'&entry='aentry

   if retrying=1 then req=req'&no_encoding=1'

   if timecheck=3 & candoadler=1 & oldfile<>'' then do
      dmd4=rx_md4(oldfile,1)
      req=req'&md4='dmd4
   end 

   if candogzip=1 & opts.!GZIP=1 & entries.mm0.!NOCE=0 then req=req'&gzip=1'
   if candorsync=1 & opts.!RSYNC=1 & entries.mm0.!NOCE=0 then do
      req=req'&rsync=1'
      rsyncok=1
   end
   else do
      rsyncok=0
   end

   if encrypts<>'' & key<>0 then do  /* shared_secret=0 means "suppress encryption */
      req=req'&encrypt='encrypts
   end 
   SHARED_SECRET=KEY

   stuff=get_a_url(opts.!server,req,auser,apwd,oldfile,,tempfile)

   stat.!bytessent=stat.!bytessent+bytessent

   if stuff='' then do 
      if notmod=1 then do               /* 304 return */
         stat.!nochange=stat.!nochange+1
         if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
         call lineout outfile,cmm0'No change (hash check) in 'tofile
         iterate
      end 
      if notmod=2 then do               /* 304 return */
         stat.!nochange=stat.!nochange+1
         if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
         if quiet<>1 then call lineout outfile,cmm0||' No change (rsync check) in 'tofile
         iterate
      end 
      stat.!noget=stat.!noget+1
      call lineout outfile,cmm0'ERROR could not GET 'afile
      if retrying=0 then do
         ir=retrys.0+1
         retrys.ir=mm0
         retrys.0=ir 
      end
      iterate
   end /* do */

   if oldfile<>"" then do               /* oldfile exists? */
     if backfile='' then do             /* then,delete it */
        fii=sysfiledelete(tofile)
     end
     else do                           /* or, move or copy oldfile to backfile */
        if translate(filespec('d',tofile))=translate(filespec('d',backfile)) then do
          fii=sysfiledelete(backfile)                   /* move old to backup */
          fii=dosrename(tofile,backfile)
          if fii<>1 then do
             call lineout outfile,fii' 'backfile'  Warning: Unable to move to backup dir: 'tofile
             foo=sysfiledelete(tofile)
          end /* do */
        end
        else do                 /* can't move, try copying */
          fii=doscopy(tofile,backfile,'R')
          if fii<>0 then do
             call lineout outfile,'Warning: Unable to copy to backup dir: 'tofile
          end /* do */
          foo=sysfiledelete(tofile)
        end
     end                        /* back file specifid */
   end                  /* oldfile exists */



/* write response to the "tofile" */
   if response_tofile=0 then do
      call write_stuff_tofile 
      if len_stuff=0 then iterate
   end 
   else do
      tempfile=stuff
      call copy_tempfile_tofile 
      if len_stuff=0 then iterate
   end 

   addo=''                      /* content encodings applied */
   if diddecrypt+didgzip+didrsync>0 then do
          addo=' < 'orisize' w/'
          if didgzip=1 then addo=addo||' GZ'
          if didrsync=1 then addo=addo||' RSY'  
          if diddecrypt=1 then addo=addo||' DeC'
    end 

   if oldfile='' then do
       stat.!created=stat.!created+1
       zz=cmm0'Creating 'tofile' ('len_stuff' bytes'||addo
       if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
       call lineout outfile,zz
   end
   else do
       zz=cmm0'Updating 'tofile' ('len_stuff' bytes'||addo
       if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
       call lineout outfile,zz
       stat.!updated=stat.!updated+1
   end
   
end

call lineout outfile,' '
call lineout outfile,'# files processed: 'tocheck
call lineout outfile,' bytes transfered: 'stat.!bytessent
call lineout outfile,'          created: 'stat.!created
call lineout outfile,'          updated: 'stat.!updated
call lineout outfile,'      not changed: 'stat.!nochange
nerrs=stat.!nocreate+stat.!noget
call lineout outfile,'           Errors: 'nerrs
if nerrs>0 then do
  call lineout outfile,'               GET failed= 'stat.!noget
  call lineout outfile,'            write problem= 'stat.!nocreate
end /* do */
call lineout outfile,' '
deletes=0
/* remove files? */
if dodelete=1 then do
   ff=destdir'\*.*'
   aa=sysfiletree(ff,'alls','FOS')
   do mm=1 to alls.0
      a1='!'||translate(strip(alls.mm))
      if didit.a1=0 then do
         deletes=deletes+1
         foo=sysfiledelete(alls.mm)
         if foo=0 then
            call lineout outfile,' deleted 'alls.mm 
         else
            call lineout outfile,' failed to delete 'alls.mm 
      end
   end 
   call lineout outfile,' Number of (orphan) files deleted: 'deletes'<p>'
end


if retrys.0>0 & retrying=0  then do  
   call lineout outfile,' '
   call lineout outfile,' ReGET 'retrys.0' unavailable files '
   call lineout outfile,' '
   retrying=1 ;tocheck=retrys.0
   i1=1  ; i2=retrys.0
   stat.=0
   signal doretry
end

/* warn client of missing info */
if retrys.0> retrys.0.!ok  then do  
   unavail=retrys.0-retrys.0.!Ok
   call lineout outfile,' '
   call lineout outfile,"Warning: "unavail" files were not successfully obtained"
   call lineout outfile,'You should re-access this UpDater package, request the "latest version",'
   call lineout outfile,'and try again.'
   call lineout outfile,' '
end

call lineout outfile,' Total elapsed time: '||time('e')||' seconds'
call lineout outfile
if asem<>0 then do
  foo=eventsem_close(asem)
  foo=eventsem_close(suicide)
end
if datatype(gosock)='NUM' then   foo=a_sockclose(gosock,1)

exit 

/*** jump here on error */
badguys:
call pmprintf("updater daemon error "Rc " at " sigl)
call lineout outfile,"updater daemon error "Rc " at " sigl
exit


/*************/
/* copy contents of tempfile to tofile-- use rexxlib move if possible
tofile should be deleted by do_backfile */

copy_tempfile_tofile:

len_stuff=0

if norexxlib=0 then do          /* use rexxlib procedures */
  if translate(filespec('d',tofile))=translate(filespec('d',tempfile)) then do
      fii=dosrename(tempfile,tofile)
      if fii=0 then DO 
          call lineout outfile,fii'  Warning: Unable to rename to: 'tofile
          RETURN 0            
      end 
   end
   else do                 /* can't move, try copying */
       fii=doscopy(tempfile,tofile,'R')
       if fii<>0 then DO 
          call lineout outfile,'  Warning: Unable to copy to: 'tempfile' 'tofile
          RETURN 0            
       end 
   END
   len_stuff=stream(tofile,'c','query size')
   return 1  
END

goo=stream(tempfile,'c','open read')
if abbrev(strip(translate(goo)),'READY')=0 then do
     call lineout outfile,fii'  Warning: Unable to open: 'tempfile
     RETURN 0            
end 
sz=charin(tempfile,'c','query size')
if sz=0 | sz='' then do
     call lineout outfile,fii'   Warning: empty file: 'tempfile
     RETURN 0            
end 

foo=stream(tofile,'c','open write')
if abbrev(translate(foo),'READY')=0 then do
       call lineout outfile,'ERROR unable to open 'tofile
       stat.!nocreate=stat.!nocreate+1
       return 0
end 

do mmw=1 to sz by 10000
   jjsize=min(10000,1+sz-mmw)
   gin=charin(tempfile,mmw,jjsize)
   foo=charout(tofile,gin)
   if foo<>0 then do
       call lineout outfile,'ERROR unable to write to 'tofile
       stat.!nocreate=stat.!nocreate+1
       return 0
   end
end 

foo=stream(tempfile,'c','close')
foo=stream(tofile,'c','close')
foo=sysfiledelete(tempfile)
len_stuff=sz
return 1


/*************/
/* write "stuff" string to tofile. tofile should be deleted by do_backfile */
write_stuff_tofile:
   len_stuff=0
   foo=stream(tofile,'c','open write')
   if abbrev(translate(foo),'READY')=0 then do
       call lineout outfile,'ERROR unable to open 'tofile
       stat.!nocreate=stat.!nocreate+1
       return 0
   end 
   aa=charout(tofile,stuff,1)
   bb=stream(tofile,'c','close')
   if aa<>0 then do
       call lineout outfile,'ERROR problem writing 'tofile
       fii=sysfiledelete(tofile)
       stat.!nocreate=stat.!nocreate+1
       return 0
   end 
   len_stuff=length(stuff)
   return 1


/********************/
/* check for unzip and rsync dlls */
check_dlls:

/* --- Load the tcp/ip function library, if necessary --- */
if RxFuncQuery("SockLoadFuncs")=1 then  do    /* not there */
  foo=RxFuncAdd( "SockLoadFuncs","rxSock","SockLoadFuncs")
  if foo=0 then call SockLoadFuncs
end
if RxFuncQuery("SockLoadFuncs")=1 then  do    /* not there */
   call lineout outfile, "ERROR: UpDater requires the rxSock dynamic link library"
   exit
end

if rxfuncquery('rx_md4')=1  then do
   foo=RXFuncAdd('RXRsyncLoad', 'RXRSYNC', 'RxRsyncLoad')
   if foo=0 then call RxRsyncLoad
end
if rxfuncquery('rx_md4')=1  then do
   candorsync=0
   candoadler=0
end
else do
   candoadler=1
end 

/* see if GZIP is available */
if candogzip=0  then return 0

if rxfuncquery('rxgzLoadFuncs')=1  then do
  foo=RXFuncAdd( 'rxgzLoadFuncs', 'rxgzLIB', 'rxgzLoadFuncs')
  if foo=0 then call rxgzLoadFuncs
end
if rxfuncquery('rxgzLoadFuncs')=1  then do
   candogzip=0
   call pmprintf("UpDater warning: unable to find rxgzlib")
end

return 0

/* ------------- */
sysmkdir2:procedure
parse arg adir

adir=strip(adir,'t','\')
ff=sysmkdir(adir)
if ff=0 then return ff

/* make the tree */
f2=adir'\'
dd=filespec('d',f2)
pp=filespec('p',f2)
if pp='\' | pp='' then return -1

pp2=strip(translate(pp,' ','\'))

do mm=1 to words(pp2)
   a1=subword(pp2,1,mm)
   a1=translate(a1,'\',' ')
   dd2=dd'\'a1
   hoo=sysmkdir(dd2)
end /* do */

return hoo

/************/
/* create a base64 packing of a message */
mk_base64:procedure

do mm=0 to 25           /* set base 64 encoding keys */
   a.mm=d2c(65+mm)
end /* do */
do mm=26 to 51
   a.mm=d2c(97+mm-26)
end /* do */
do mm=52 to 61
   a.mm=d2c(48+mm-52)
end /* do */
a.62='+'
a.63='/'

parse arg mess
s2=x2b(c2x(mess))
ith=0
do forever
   ith=ith+1
   a1=substr(s2,1,6,0)
   ms.ith=x2d(b2x(a1))
   if length(s2)<7 then leave
   s2=substr(s2,7)
end /* do */
pint=""
do kk=1 to ith
    oi=ms.kk ; pint=pint||a.oi
end /* do */
j1=length(pint)//4
if j1<>0 then pint=pint||copies('=',4-j1)
return pint


/*************/
/* get ip address of this machine */
get_hostname:procedure
dot_addr = SockGetHostId()
rc = SockGetHostByAddr(dot_addr, "host.")
return host.name

/*******************************************/
rsync_synopsis:procedure expose outfile

parse arg afile,nblocks

if nblocks='' then nblocks=45
if datatype(nblocks)<>'NUM' then nblocks=45
if nblocks<10 | nblocks>255 then nblocks=45   /* 255 limit on # of blocks */

if afile='' then return "ERROR no old-version file specified"

/* read "Afile" */
aa=translate(stream(afile,'c','open read'))
if  abbrev(aa,'READY')=0 then return "ERROR could not open "afile
isize=stream(afile,'c','query size')
if isize='' | isize=0 then do
    call pmprintf('ERROR 'afile " is unaccessible")
    return ''
end
astuff=charin(afile,1,isize)
aa=stream(afile,'c','close')


blocksize=trunc(0.9999 + (isize/nblocks))
if blocksize<200 then do
    blocksize=200
    nblocks=trunc((isize/blocksize)+0.999)
end /* do */
ac1=d2c(blocksize)
ac1=right(ac1,4,x2c('00'))
ac1=ac1||d2c(nblocks)
iat=1
do mm=1 to nblocks
  if mm=nblocks then
     ablock=substr(astuff,iat)
  else
     ablock=substr(astuff,iat,blocksize)
  ac0=left(x2c(rx_rsync32_md4(ablock)),8)
  ac1=ac1||ac0
  iat=iat+blocksize
end
ac1=mkpack64(ac1)

return ac1

/**********************/
mkPACK64:procedure expose outfile
parse arg mess

biga=xrange('A','Z')||xrange('a','z')||xrange('0','9')||'+/'

s2=x2b(c2x(mess))

nith=trunc((length(s2)/6)+.9)
cont=copies(' ',nith)
oof=""
do mm=0 to 63
      oof=oof||x2c(b2x(right('00'||x2b(d2x(mm)),6)))
end /* do */
do ith=1 to nith 
  a1=substr(s2,(ith*6)-5,6,0)
  cont=overlay(x2c(b2x(a1)),cont,ith) 
end /* do */
pint=""
pint=translate(cont,biga,oof)
j1=length(pint)//4
if j1<>0 then pint=pint||copies('=',4-j1)
return pint




/***********/
/* adler-32 crc of file aname */
fileadler:procedure expose candorsync outfile
parse arg aname

if candorsync=0 then return ''

foo=stream(aname,'c','open read')
if abbrev(strip(translate(foo)),'READY')=0 then return ''

sz=stream(aname,'c','query size')
if sz='' | sz=0 then return ' '
aa=charin(aname,1,min(100000,sz))
adler=rx_adler32(aa)
igot=100000
do forever
      if igot>=sz then leave
      iget=min(100000,sz-igot)
      aa=charin(aname,,iget)
      adler=rx_adler32(aa,adler)
      igot=igot+100000
end
foo=stream(aname,'c','close')

return adler

/************************/
replacestrg:procedure

exactmatch=0
backward=0 ; doall=0

parse arg astring ,  target   , putme , type , exactmatch

type = translate(type)
if type="BACKWARD" then backward="YES"
if type="ALL" then doall="YES"

iat=1
joelen=length(target)
joelen2=length(putme)

doagain:                /* here if doall=yes */
 if exactmatch="YES" then do
    if   backward="YES" then
        joe= lastpos(target,astring)
    else
        joe= pos(target,astring,iat)
 end
 else do
   if   backward="YES" then
        joe= lastpos(translate(target),translate(astring))
    else
        joe= pos(translate(target),translate(astring),iat)
 end
 if joe=0 then
         return astring

 astring=delstr(astring,joe,joelen)
 if putme<>' ' then
    astring=insert(putme,astring,joe-1)

 if doall="YES" then do
     iat=joe+joelen2
     signal doagain
 end
/* else, all done */
 return astring

/***************************************************************************/
/**********************************************************/
/***********/
/* get a url, using socket calls */
get_a_url:procedure expose servername  rsyncok verbose notmod retrys. temp_dir encrypt. ,
                   shared_secret listfile opts. outfile response_tofile ,
                   norexxlib  timeout mm0a bytessent , 
                   didrsync didgzip orisize diddecrypt askupwd  gosock opts.
parse arg server,request,username,pwd,oldfile,forcesock,tempfile

crlf = '0d0a'x
sorig_size=''
ith=0
bytessent=0

arsyncok=rsyncok
askupwd=0
req0=request
iter=0
notmod=0

if forcesock=1 | gosock='' then do
   gosock=sockhttp_opensock(gosock,server)
   if abbrev(gosock,'-1')=1 then do
      call pmprintf("UpDaemon Error, unable to create socket to " server)
      gosock=''
     exit
  end
end

agin:           /* jump here if problem (and try GET without encoding */

didrsync=0; didgzip=0 ;diddecrypt=0
response_tofile=0

message='GET 'request ' HTTP/1.1'crlf'HOST: 'server||crlf
message=message||'Referer: UpDaemon@'||servername||crlf
if forcesock=1 then message=message||'Connection: close' crlf /* forcesock implies close */

if arsyncok=1 & oldfile<>'' then do
   rsign=rsync_synopsis(oldfile)
   if rsign<>'' then  message=message||'Rsync-signature_updater: 'rsign||crlf
end

if username<>'' then do
   upwd=mk_base64(strip(username)':'strip(pwd))
   upwd='Basic 'upwd
   message=message||'Authorization: '||upwd||crlf
end
message=message||crlf
do oppie=1 to 2

   rc = SockSend(gosock, message)   /* SEND THE REQUEST TO THE SERVER */

/* !!! Now get the response !!!!! */
  got=sockhttp(gosock,timeout,tempfile,10000)  /* WAIT FOR, AND GET, RESPONSE */

  parse var got statline '0d0a'x got              /* get the status */
  parse var statline asockstat sockinfo 
  asockstat=strip(asockstat)
  select                        /* CHECK FOR ERROR CONDITIONS */
    when asockstat=-1 then  do
      call pmprintf('UpDaemon ERROR  Unable to complete request: 'sockinfo','gosock)
      if datatype(gosock)='NUM' then   foo=a_sockclose(gosock,1)
      gosock=''
      return ''
    end
  
    when asockstat=-2 & ith=1 then do
         call pmprintf('UpDaemon ERROR  Unable to connect to 'server)
         if datatype(gosock)='NUM' then foo=a_sockclose(gosock,2)
         gosock=''
         return ' '
    end

    when asockstat=-2 then do
      ith=1
      if verbose>1 then call pmprintf('UpDaemon warning Can not reuse ' gosock', creating new socket ')
      gosock=sockhttp_opensock(gosock,server,httpport)
      if abbrev(gosock,'-1')=1 then do
         call pmprintf('UpDaemon warning error connecting: 'gosock)
         gosock=''
         return ' '
      end 
      iterate           /* try again, with freshly minted socket */
    end

    otherwise leave             /* good response */
   end                  /* select */
end                     /* oppie */

/* IF here, response recieved. Check for non 200 response codes */
parse var sockinfo sockclosed respcode . ; respcode=strip(respcode)

select
  when respcode=401 then do
     askupwd=1
     if forcesock=1 then  foo=a_sockclose(gosock,3)
     return ' '
  end 
  when respcode>399  then do
     call pmprintf('ERROR not able ('respcode') to query 'server)
     if forcesock=1 then  foo=a_sockclose(gosock,4)
     return ' '
  end
  when respcode=304 then do            /* not modified */
      notmod=1
      if forcesock=1 then  foo=a_sockclose(gosock,5)
      return ''
   end 
   when respcode=200 then do
      nop
   end 
   otherwise do
     call pmprintf('UpDaemon ERROR Unknown response code ('respcode') to 'server)
     if forcesock=1 then  foo=a_sockclose(gosock,24)
     return ' '
   end  
end

/* if here, got a 200 response code */
if asockstat=1 then do  /* body is in tempfile */
   gotfile=1
   t2=got
   stuff=tempfile
end 
else do                 /* returned result has headers and body */
   gotfile=0
   findit=crlf||crlf
   ifindit=pos(findit,got)
   t2=substr(got,1,ifindit)
   stuff=substr(got,ifindit+4)             /* THE ENTITY BODY */
end

/* look for varions response headers */
cencoding='';xupdater=''
clen=''

do until t2=''
   parse var t2 ahdr (crlf) t2
   parse upper var ahdr aname ':' avalue
   if strip(aname)='CONTENT-ENCODING' then cencoding=strip(avalue)
   if strip(aname)='X-UPDATER' then xupdater=strip(avalue)
   if strip(aname)='CONTENT-LENGTH' then clen=strip(translate(avalue,' ','0d0a0009'x))
   if strip(aname)='X-NOTMOD' then do
       otmod=strip(translate(avalue,' ','0d0a0009'x))
       if otmod=1 then do
          notmod=2
          if forcesock=1 then  foo=a_sockclose(gosock,6)
          return ''
        end
   end
end /* do */

if abbrev(translate(xupdater),'ERROR')=1 then do 
  if verbose>1 then call pmprintf("ERROR: "||'0d0a'x||got)
  if forcesock=1 then  foo=a_sockclose(gosock,7)
  return ''  /* some kind of error, give up */
end /* do */

if xupdater<>'' then do
  parse upper var xupdater xup 'MD4=' smd4
  parse upper var xup . 'SIZE=' sorig_size . ',' .
  if datatype(sorig_size)<>'NUM' then sorig_Size=''
end
if gotfile=0 then
   orisize=length(stuff)
else
   orisize=stream(tempfile,'c','query size')

bytessent=orisize


if clen<>''  then do
 if orisize<>clen then do  /* error, not correct amount of bytes recieved */
   call pmprintf('UpDaemon ERROR content-length mismatch: 'clen ', actual='||orisize)
   if forcesock=1 then  foo=a_sockclose(gosock,8)
   return ' '
 end 
end

/* unencode response? */
encodeok=1
badenc=''
if cencoding<>'' then do   /* unencode from right to left */
    cencoding=translate(cencoding,' ',',')

    do jj=words(cencoding) to 1 by -1
        aword=strip(word(cencoding,jj))
        select
           when abbrev(aword,'ENCRYPT=')=1 then do
              if gotfile=1 then do
                  tempfile=do_decrypt(aword,tempfile,1)  /* tempfile will be name of decrypted file */
                  stuff=tempfile
              end
              else do
                 stuff=do_decrypt(aword,stuff)
              end
              if stuff='' then do
                if verbose>1 then call pmprintf('UpDaemon error: unable to decrypt 'request)
                if forcesock=1 then  foo=a_sockclose(gosock,9)
                return ""
              end
              diddecrypt=1
           end 
           when aword='GZIP' then do
              if gotfile=0 then do
                 stuff=do_ungzip(stuff)
              end
              else do
                 tempfile=do_ungzip(tempfile,1)
                 stuff=tempfile
              end
              if stuff='' then do
                if verbose>1 then call pmprintf("UpDateMe error: bad ungzip")
                encodeok=0
                leave
              end
              didgzip=1
           end
           when aword='GDIFF' then do
              if oldfile='' then do             /* should never happen */
                encodeok=0
                leave
              end /* do */

              if gotfile=1 then do               /* differences are in a file */
                   tempfile2=make_tempname('UD3',norexxlib,temp_dir)
                   stat=rsync_ungdiff_file(oldfile,tempfile,smd4,tempfile2)
                   if stat='' then do
                      if verbose>1 then call pmprintf('UpDaemon error: bad ungdiff ')
                      encodeok=0
                      leave
                   end
                   foo=sysfiledelete(tempfile)
                   tempfile=tempfile2
                   stuff=tempfile
              end
              else do                           /* differences are in a string */
                 stuff=rsync_ungdiff(oldfile,stuff,smd4)
                 if stuff='' then do
                    if verbose>1 then call pmprintf('UpDaemon error: bad ungdiff '  )
                    encodeok=0
                    leave
                 end
              end
              didrsync=1
           end                   /*  GDIFF encoding */
           when aword='IDENTITY' then iterate
           otherwise do
                badenc=badenc' 'aword
                encodeok=0
                leave
           end
        end  /* select */
    end /* cencoding */
end /* cencoding <> '' */

if encodeok=0 then do           /* problem encoding-- retrieve without any encoding ? */
    if iter=0 then do
        areq=''
        do until request=''
            parse var request a1 '&' request ; ta1=translate(a1)
            if abbrev(ta1,'GZIP')+abbrev(ta1,'RSYNC')=0 then areq=areq||'&'||a1
        end 
        request=strip(areq,,'&') ;iter=1 ;arsyncok=0; 
        signal agin
    end /* do */
    else do
        if verbose>0 then 
          call pmprintf('UpDaemon ERROR unknown encoding encountered ('badenc') trying w/o encoding')
    end /* do */
end /* do */

if badenc<>'' then do
      call pmprintf('UpDaemon ERROR unknown encoding 'badenc)
      if forcesock=1 then  foo=a_sockclose(gosock,10)
      return ""
end

/* check final size */
if sorig_size<>'' then do
  if gotfile=0 then
         len_final=length(stuff)
  else
         len_final=stream(tempfile,'c','query size')
  if strip(sorig_size)<>len_final then do
       call pmprintf('UpDaemon ERROR size ('len_final') does not match original size ('sorig_size')')
       if forcesock=1 then  foo=a_sockclose(gosock,11)
       return ""
  end 
end 

if forcesock=1 then  foo=a_sockclose(gosock,12)
if gotfile=1 then response_tofile=1
return stuff



/***************************************************/
/*  rsync_ungdiff: undifference, using a gdiff difference file
                   (that may have have been produced by rsync_gdiff

     status=rsync_ungdiff(oldver_file,diff_file,newfile,GDMD4,verbo)
        OLDVER_FILE -- old version of the file
        diff_file  -- the gdiff file (in a string)
        gdmd4 -- md4 of "server side" new file -- use as an error check
                 if not specified, then error check will not be done


**************************************/

rsync_ungdiff:procedure expose   norexxlib  mm0a listfile
  parse arg afile,s1,smd4


crlf='0d0a'x

/* read "Afile" */
aa=translate(stream(afile,'c','open read'))
if  abbrev(aa,'READY')=0 then do
  call pmprintf('UpDaemon ERROR could not open 'afile)
  return ''
end
iasize=stream(afile,'c','query size')
if iasize='' | iasize=0 then do
    call pmprintf('UpDaemon ERROR unable to apply rsync to 'afile)
    return ''
end
astuff=charin(afile,1,iasize)
aa=stream(afile,'c','close')

/* start building, according to the gdiff format */

/* check for gdiff header (char value of 'd1ffd1ff'x, and version # i.e; '04'x)*/
parse var s1 gheader +4 foover +1  s1
ggh=x2c('d1ffd1ff')
if gheader<>ggh then do
    call pmprintf('UpDaemon Error in difference file: not a gdiff file: 'c2x(gheader))
    return ''
end

/* ignore version number */
bstuff=''
noted=0
do forever
   if length(s1)=0 | s1='' then do
       call pmprintf('UpDaemon ERROR in RSYNC ')
       return ''
   end
   
   parse var s1 atype +1 s1
   itype=c2d(atype)
   select
      when itype=0 then leave   /* 0 signals eof */
      when itype=255 then return 'ERROR gdiff copy operation too large'
      when  itype>0 & itype < 247 then do  /* get itype bytes */
         parse var s1 ccs +(itype) s1
         bstuff=bstuff||ccs
      end /* do */
      when itype=247 then do    /* get <65k bytes */
          parse var s1 jget +2 s1
          jget=c2d(jget)

          parse var s1 ccs +(jget) s1
          bstuff=bstuff||ccs
      end /* do */
      when itype=248 then do    /* get < 2billion bytes */
          parse var s1 jget +4 s1
          jget=c2d(jget)

          parse var s1 ccs +(jget) s1
          bstuff=bstuff||ccs
      end /* do */
      otherwise do   /* copy range of bytes */
         select
           when itype=249 then parse var s1 istart +2 iget +1 s1  
           when itype=250 then parse var s1 istart +2 iget +2 s1
           when itype=251 then parse var s1 istart +2 iget +4 s1
           when itype=252 then parse var s1 istart +4 iget +1 s1
           when itype=253 then parse var s1 istart +4 iget +2 s1
           when itype=254 then parse var s1 istart +4 iget +4 s1
           otherwise return 'ERROR: impossible code in gdiff '
         end
         istart=c2d(istart); iget=c2d(iget)

         bstuff=bstuff||substr(astuff,istart,iget)
      end               /* otherwise range */
   end                  /* gdiff codes */
end                     /* parsing gdiff file */

/* compute md4 of this constructed file */
if smd4<>'' then do 
  b2md4=rx_md4(bstuff)
  if strip(translate(b2md4)) <> strip(translate(smd4)) then do
     call pmprintf('UpDaemon ERROR md4 does not match: 'b2md4', 'smd4)
     return ''
  end /* do */

end

return bstuff

/***************************************/
/***** call this if "all file" ungdiff */

rsync_ungdiff_file:procedure expose   norexxlib  mm0a listfile ,
                outfile  temp_dir

parse arg oldfile,difffile,smd4,tempfile

crlf='0d0a'x

/* open for read "oldfile" */
aa=translate(stream(oldfile,'c','open read'))
if  abbrev(aa,'READY')=0 then do
  call pmprintf('UpDaemon ERROR could not open 'oldfile)
  return ''
end
oldsize=stream(oldfile,'c','query size')
if oldsize='' | oldsize=0 then do
    call pmprintf('UpDaemon ERROR empty oldfile: 'oldfile)
    return ''
end

/* open for read "difffile" */
aa=translate(stream(difffile,'c','open read'))
if  abbrev(aa,'READY')=0 then do
  call pmprintf('UpDeamon ERROR could not open 'difffile)
   foo=stream(oldfile,'c','close')
  return ''
end
diffsize=stream(difffile,'c','query size')
if diffsize='' | diffsize=0 then do
    call pmprintf('UpDaemon ERROR empty diff file: 'difffile)
    foo=stream(oldfile,'c','close')
    return ''
end
difffile_left=diffsize ; diffbuffer=''

/* check for gdiff header (char value of 'd1ffd1ff'x, and version # i.e; '04'x)*/
aa=rsync_pull_diff(5,1)
if aa='' then return ''

parse var aa gheader +4 .
ggh=x2c('d1ffd1ff')                     /* ignore version number */
if gheader<>ggh then do
    call rsync_error 'ERROR in difference file: not a gdiff file: 'c2x(gheader)' 'gheader','difffile
    return ''
end

/* start building, according to the gdiff format */
do forever

   atype=rsync_pull_diff(1)
   IF atype='' then return ''
   itype=c2d(atype)

/* process, according to the gdiff code */
   select

      when itype=0 then leave   /* 0 signals eof */
      when itype=255 then do
        call rsync_error  'ERROR gdiff copy operation too large'
        return ''
      end
      when itype < 247 then do  /* get itype (<247) bytes */
         ccs=rsync_pull_diff(itype)
         if ccs='' then return ''
         foo=charout(tempfile,ccs)
         if foo<>0  then do
            call rsync_error 'ERROR writing (a) to 'tempfile
            return ''
         end 
      end /* do */
      when itype=247 then do    /* get <65k bytes from  */
          jget=rsync_pull_diff(2)
          if jget='' then return ''
          jget=c2d(jget)
          ccs=rsync_pull_diff(jget)
          if ccs='' then return ''
          foo=charout(tempfile,ccs)
          if foo<>0  then do
             call rsync_error 'ERROR writing (b) to 'tempfile
             return ''
          end 
      end /* do */
      when itype=248 then do    /* get < 2billion bytes */
          jget=rsync_pull_diff(4)
          if jget='' then return ''
          jget=c2d(jget)
          do mm=1 to jget by 10000
             jget2=min(10000,1+jget-mm)
             ccs=rsync_pull_diff(jget2)
             if ccs='' then return ''
             foo=charout(tempfile,ccs)
             if foo<>0  then do
                call rsync_error 'ERROR writing (c) to 'tempfile
                return ''
             end 
          end           /* to jget by 10000 */
      end /* do */
      otherwise do   /* copy range of bytes from old file */
         
         select
           when itype=249 then do
              s1=rsync_pull_diff(3)
              if s1='' then return ''
              parse var s1 istart +2 iget +1 
           end
           when itype=250 then do
              s1=rsync_pull_diff(4)
              if s1='' then return ''
              parse var s1 istart +2 iget +2 
           end
           when itype=251 then do
              s1=rsync_pull_diff(6)
              if s1='' then return ''
               parse var s1 istart +2 iget +4
           end
           when itype=252 then do
              s1=rsync_pull_diff(5)
              if s1='' then return ''
              parse var s1 istart +4 iget +1 
           end
           when itype=253 then do  
               s1=rsync_pull_diff(6)
              if s1='' then return ''
               parse var s1 istart +4 iget +2
           end
           when itype=254 then do
              s1=rsync_pull_diff(8)
              if s1='' then return ''
              parse var s1 istart +4 iget +4 
           end
           otherwise return 'ERROR: impossible code in gdiff '
         end                    /* 249 to 254 codes */

         istart=c2d(istart); iget=c2d(iget)
         iend3=istart+iget-1
         if iend3>oldsize  then do
             call rsync_error 'ERROR read past end of reference-diff file 'oldfile
             return ' '
         end /* do */
         do mm=istart to iend3 by 10000
             nget=min(10000,1+iend3-mm)
             ccs=charin(oldfile,mm,nget)
             foo=charout(tempfile,ccs)
             if foo<>0  then do
                call rsync_error 'ERROR writing (d) to 'tempfile
                return ''
             end 
         end           /* to jget by 10000 */

      end               /* otherwise get from oldfile */
   end                  /* gdiff codes */

end                     /* parsing gdiff file */

foo=stream(difffile,'c','close')
foo=stream(oldfile,'c','close')
foo=stream(tempfile,'c','close')

/* compute md4 of this constructed file */
if smd4<>'' then do 
  b2md4=rx_md4(tempfile,1)
  if strip(translate(b2md4)) <> strip(translate(smd4)) then do
     call pmprintf('UpDeamon ERROR md4 does not match: 'b2md4', 'smd4)
     return ''
  end 
end

return tempfile


/*** rsync error handler ******/
rsync_error:
parse arg amess
if amess<>'' then call pmprintf('UpDaemon Error' amess)
foo=stream(difffile,'c','close')
foo=stream(oldfile,'c','close')
return ''


/************/
/* pull stuff from difffile, or from difffile's buffer */
rsync_pull_diff:procedure expose diffbuffer difffile difffile_left   outfile 
parse arg nget,i1


do forever              /* may need to dig stuff out of file a few times */

lendiff=length(diffbuffer)
if lendiff>=nget then do
   parse var diffbuffer ccs +(nget) diffbuffer
   if lendiff=nget then diffbuffer=''
   return ccs                           /* success, return nget bytes */
end

/* else, add to diffbuffer */
if difffile_left<1 then do               /* error, buffer can't be filled*/
   call rsync_error nget' ERROR premature end of diff file 'difffile 
   return ''
end 

if i1=1 then
 t1=charin(difffile,i1,min(difffile_left,10000))
else
 t1=charin(difffile,,min(difffile_left,10000))


difffile_left=difffile_left-length(t1)

diffbuffer=diffbuffer||t1

end             /* compare against nget again */




/* ---------------------------------- */
/* return 1 if adir is an existing (possibly empty) directory , 0 if not */
ais_dir:procedure expose norexxlib    mm0a listfile
parse arg adir

if norexxlib=0 then do  /* use rexxlib, if available */
   ii=dosisdir(adir)
   return ii
end /* do */

adir=strip(adir)
adir=strip(adir,'t','\')
nowdir=directory()
nowdrive=filespec('d',nowdir'\')
nowpath=filespec('p',nowdir'\')
adr=filespec('d',adir)
if adr='' then do
   if abbrev(adir,'\')=0 then 
       adir=nowdrive||nowpath||adir
   else
       adir=nowdrive||adir
end /* do */

foo=sysfiletree(adir,goo,'D')
if  goo.0>0  then return 1
return 0



/***********************************/
/* decrypt a file */
do_decrypt:procedure expose  temp_dir shared_secret  encrypt. listfile  outfile ,
                          norexxlib opts.
parse arg arg1,astring,isfile

useenc=''
parse var arg1 . '=' a2 ; a2='!'||strip(translate(a2))
if symbol('ENCRYPT.'||a2)<>'VAR' then return ''  /* not supported encrytion type */
doit=encrypt.a2

if shared_secret='' then do
    shared_secret=ask_shared()
end 
shared_secret=translate(shared_secret)

tmpf=make_tempname('1',norexxlib,temp_dir)  /* where the decrypted stuff will be written to */

/* if file, then astring is name ofenocded file */
if isfile<>1 then do
   tmpfenc=make_tempname('ENC',norexxlib,temp_dir)
end /* do */

foo=sysfiledelete(tmpf)
if isfile<>1 then do
  foo=sysfiledelete(tmpfenc)
  joo=charout(tmpfenc,astring,1)
  oog=stream(tmpfenc,'c','close')
  if joo<>0 then do
    foo=sysfiledelete(tmpf)
    foo=sysfiledelete(tmpfenc)
    return ''
  end 
end
else do
  tmpfenc=astring
end

doit=replacestrg(doit,'%IN',tmpfenc,'ALL')
doit=replacestrg(doit,'%OUT',tmpf,'ALL')
doit=replacestrg(doit,'%KEY',shared_secret,'ALL')
doit=strip(doit)
aa=word(doit,1)
signal on syntax name noenc ; signal on  error name noenc

if pos('(',aa)>0 then do
    doit='foo='doit
    interpret doit 
end
else do
   address cmd '@'doit 
end /* do */

signal on error name badguys; signal on syntax name badguys
foo=sysfiledelete(tmpfenc)     /* don't need this anymore */

if isfile=1 then do
   return tmpf          /* decrypted file is now the working file */
end /* do */


oof=strip(translate(stream(tmpf,'c','open read')))
if abbrev(oof,'READY')=0 then do
  foo=stream(tmpf,'c','close')
  foo=sysfiledelete(tmpf)
  return ''         /* some kind of error */
end
sz=stream(tmpf,'c','query size')
if sz=0 | sz='' then do
  foo=stream(tmpf,'c','close')
  foo=sysfiledelete(tmpf)
  return ''         /* some kind of error */
end
aa=charin(tmpf,1,sz)
foo=stream(tmpf,'c','close')
foo=sysfiledelete(tmpf)
return aa



noenc:          /* jump here on error in decoding software */
signal on error name badguys; signal on syntax name badguys
foo=sysfiledelete(tmpf)
foo=sysfiledelete(tmpfenc)
return ''



/*********************************/
/* find a name for a temporary file */
make_tempname:procedure
parse arg ext,norexxlib,tempdir
ext=strip(ext) ; tempdir=strip(tempdir)

if pos(norexxlib,'01')=0 then do
   norexxlib=rxfuncquery('REXXLIBREGISTER')
end 

if tempdir='' then tempdir=value('TEMP',,'os2environment')

if norexxlib<>0 then
    tfoo='t'||right(date('d')||time('s'),7,'0')
else
    tfoo='t'||right(dospid()||dostid(),7,'0')

tempfile=tempdir'\'tfoo'.'||strip(ext,'l','.')
foo=stream(tempfile,'c','open write')
if abbrev(strip(translate(foo)),'READY')=0 then return ''      /* unable to write to this file */
foo=stream(tempfile,'c','close')
foo=sysfiledelete(tempfile)             /* delete it (possibly 0 length file just created */
if stream(tempfile,'c','query exists')<>'' then return '' /* could not delete */
return tempfile         /* success! */


/***********************************/
/* un gzip a file */
do_ungzip:procedure  expose temp_dir listfile  outfile   norexxlib
parse arg astring,isfile

if isfile=0 | isfile='' then do
   a2=rxgzInflateString(astring)
   return a2                            /* return as string */
end

tempfile=make_tempname('UD2',norexxlib,temp_dir)
a2=rxgzInflateFile(astring,tempfile)
if a2<>0 then return ''
foo=sysfiledelete(astring)

return tempfile               /* tempfile is now working file */



/***************************************************************************/
/* SOCKhttp: a replacement for a sequence of sockrecv calls; 
             optimized for http requests 

  call as

      stuff=sockhttp(socket,[socket_timeout,tfile,max_notfile])

  where:
      socket == a socket that's been established using sockhttp_opensock
                (or with sockconnect).
               Alternatively, an ip address in the form
                        ii.jj.kk:port
                where
                        ii.jj.kk is an ip address
                        :port is an optional port (:80 is the default)
                If this address is provided, then a socket will be opened, 
                used, and then closed.

                Note: if a socket is provided, it will NOT be closed 
                      (you should close it using stat=a_sockclose(socket)

    socket_timeout == optional. a timeout value in seconds 
                      (per socket call, default=30)
    tfile    == optional. file to write results to (default, do NOT write to file)
 max_Notfile == optional. maximum response size for writing to string. If response
                > then this, then return headers and put body into tfile
                If 0 (the default), then always write to file 
and
   stuff =  status_line '0d0a'x response

where: 

  status_line: a space delimited set of values:
      stat_code   message

  stat_code: takes values
     -2 = error, probably due to specified socket not being open
     -1 = error 
      0 = success, stuff contains entire response line, response headers & body
      1 = success, stuff contains response line & response headers,
         resonse body will be in tfile.
         This can only happen if  stat tfile<>'' & length(response)>max_notfile 

  message:
     if stat_code<0, then a several word description of the error
     otherwise, wasclosed response_line; where--
        wasclosed: 0 if connection was not closed, 1 if it was
        response_code: the response code (i.e.; 200)


 response:
   if stat_code<0, then stuff will be empty
      stat_code=0, the response headers and body
      stat_code=1, the headers (body will be in tfile)

Example:
   stuff=sockhttp(,,'foo.tmp')
   parse var stuff statline '0d0a'x stuff1
   parse var statline statcode message
   select 
     when statcode<0 then call pmprintf('UpDaemon Error: "message)
     when statcode=0 then do
          parse message wasclosed response_code
          parse var stuff1 response_headers '0d0a0d0a'x response_body
      end
     when statcode=1 then do
       response_headers=stuff1   / * response_body is in foo.tmp * /
     end
     otherwise
   end


Notes:
    * if tfile not defined, then max_notofile is ignored
    * see "user changable parameters" for some default values
    * socket_timeout refers to maximum seconds between "sockrecv" requests. 
      It does NOT refer to total length of time required to recieve a message.
      Thus, a number of medium length delays (say, a few seconds required
      to complete each of several sockrecv requests) will NOT cause a timeout 
      (in other words, the timeout counter is reset upon each successful 
      completion of a 1000 byte sockrecv request).
  
*/

SOCKhttp:PROCEDURE expose  outfile mm0a verbose

/* ------------ User changable parameters  --------------------- */

/* default time out in seconds for each set of socket calls
   (if timeout argument not given) */
def_socket_timeout=30

/* default value of max_notfile */
def_max_notfile=0       /* if tfile specified, 0 means "always write body to tfile*/

/* maximum length of response. If response > than this, an error is returned */
maxlen=1000000000  /* 1 gigabyte */


/* ------------ end of user changable parameters -----------------*/

PARSE Arg socket,socket_timeout,tfile,max_notfile                    

if datatype(socket_timeout)<>'NUM' then socket_timeout=def_socket_timeout
if datatype(socket_timeout)<>'NUM' then do
   return '-1 Bad socket timeout value : 'socket_timeout
end 

if datatype(max_notfile)<>'NUM' then max_notfile=def_max_notfile
if datatype(max_notfile)<>'NUM' then do
   return '-1 Bad max_notfile value : 'max_notfile
end 

if maxlen='' then maxlen=1000000000
if datatype(maxlen)<>'NUM' then do
   return '-1 Bad maxlen value : 'maxlen
end 

isclosesock=0
if pos(".",socket)>0 then do             /* socket is an ip address */
   parse var socket anip ':' aport ; aport=strip(aport) ;anip=strip(anip)
   if aport=''  then aport=80
   socket=sockhttp_opensock(,anip,aport)
   if abbrev(socket,'-1')=1 then return '-1 Unable to create socket '
   isclosesock=1
end /* do */



/* READY to start recieving */

if Sockioctl(socket,'FIONBIO',1) = -1 then  do      /* switch to nonblocking mode */
   Return '-1 crashed in sockioctl 'socket':'errno
end

/* global parameters */
incoming=''
fileon=0
rheaders=''
clen='' ; ischunked=0
len_toget=''
wasclosed=0
trailers=''

ict=0

/* first, get header */
call sockhttp_headers    /* also sets ichunked, incoming, and clen */

ares=result
if ares<>'' then do
  if isclosesock=1 then do
      foo=a_sockclose(socket,13)
  end
  return ares    /* ares might be an error message, starting with -1 */
end

parse var rheaders respline '0d0a'x response
parse var respline . icode .

if icode>299 & icode<400 then do        /* 300, hence no body */
   return '0 0 'icode||'0d0a'x||rheaders
end

/* 3 ways of sending data
1) with content-length header
2) using chunked encoding
3) closing connection at end of data transferal */

if ischunked<>1 then do
   if len_toget=0 then
        nop             /* incoming was gotten */
   else
        call sockhttp_nonchunk
end
else do
   call sockhttp_chunk
   if trailers<>'' then rheaders=rheaders||'0d0a'x||trailers
end

/* did an error occur? */
ares=result
if ares<>'' then do
  if isclosesock=1 then foo=a_sockclose(socket,14)
  return ares    /* ares might be an error message, starting with -1 */
end

if sockioctl(socket,'FIONBIO',0) = -1 then do          /* switch to blocking mode */
    if verbose>0 then call pmprintf('UpDaemon Sockhttp: ioctl error on switch to blocking mode: ' errno)
end

daclosed=0

if wasclosed=1 | isclosesock=1 then do
  daclosed=1 
  foo=a_sockclose(socket,15)
end

if fileon=0 then 
   return '0 'wasclosed' 'icode||'0d0a'x||rheaders||'0d0a0d0a'x||incoming
else
   return '1 'wasclosed' 'icode||'0d0a'x||rheaders


/***** ----------------- ***/
/* read header section (and response line)
  Sets:  rheaders, incoming, ischunked, clen 
if no error, return ''. Otherwise, return '-1 error message */

sockhttp_headers:

i1000=600
timeout=socket_timeout
Do While TimeOut > 0

    res=Sockrecv(socket,'data',i1000) 

    if res=0 then do
       return '-1 Socket prematurely closed'
    end /* do */

    if res<0 then do                           /* error condition ? */
      If errno <> 'EWOULDBLOCK' THEN  do      /* real crash ? */
          if errno='ECONNRESET' | errno='ENOTSOCK' then  do
              Return '-2 socket closed?  '  /* might be bad socket */
          end
          else do
              Return '-1 crashed in sockrecv 'errno  /* yes, real crash */
          end
      end

      /* not a crash, but no data available:  errno = EWOULDBLOCK & sockrecv returned -1 */

      ict=ict+1
      goo=syssleep(0.25)           /* relinquish the CPU for one second, this is from REXXUTIL */

      TimeOut = TimeOut - 0.25 /* count down my timer */
      Iterate                    /* loop again */
   End                           /* if sockrecv = -1 */
   if res<0 then return "-1 Socket timed out " 
   ikk=pos('0d0a0d0a'x,data)

   if ikk=0 then do                /* all of DATA is header info */
      rheaders=rheaders||data
      iterate                   /* sockrecv next block */
  end

/* if here, got all of the headers */
  rheaders=rheaders||left(data,ikk-1)

  incoming=substr(data,ikk+4)          /* get beginning of content */

/* find content-length, or "chunked" transfer-encoding */
   rhh=rheaders
   do until rhh=''
       parse var rhh aline '0d0a'x rhh
       parse upper var aline aname ':' aval ;
       aname=strip(translate(aname)) ; aval=strip(translate(aval))
       select 
           when aname='CONTENT-LENGTH' then clen=aval
           when aname='TRANSFER-ENCODING' then do
               if pos('CHUNKED',aval)>0 then ischunked=1
           end
           otherwise nop
       end
    end             /* found end of response headers */
    if clen<>'' then do
        len_toget=clen-length(incoming)
    end
    leave
end                 /* building rheaders */
return ''               /* blank return signals okay */

/**************************/
/* chunked read of content   sets INCOMING, WASCLOSED, FILEON; writes to TFILE */

sockhttp_chunk:
buffer=incoming         /* left over from header read */
i1000=1000
incoming=''

totbytes=0
if verbose>0 then call charout,'Bytes read: 'totbytes||'0d'x

do forever              /* come here to read another chunk */

   if verbose>0 then call charout,'Bytes read: 'totbytes||'0d'x

   if length(incoming)>max_notfile & tfile<>'' then do
        call write_to_tfile
   end 

  ipos=pos('0d0a'x,buffer)
  if ipos=0 then do                  /* recharge buffer */
     data=sockin(socket,i1000,socket_timeout,verbose)
     parse var data scode '0d0a'x buffer0 

     if scode<0 then return data   /* some kind of error */
     if scode=0 | buffer0='' then return '-1 {a} premature close of chunked response'
     buffer=buffer||buffer0
     totbytes=totbytes+length(buffer0)
     iterate
  end      

/* determine size of chunk */
  aa=left(buffer,ipos-1)
  buffer=substr(buffer,ipos+2)
  parse var aa gg ';' .
  gtsize=x2d(strip(gg))

  if gtsize=0 then leave               /* no more content to get */
  if gtsize<0 then return '-1 error reading chunk size '

/* is this chunk||crlf within the buffer */
  if (gtsize+2)<=length(buffer) then  do
     incoming=incoming||left(buffer,gtsize)
     if gtsize+2=length(buffer) then
        buffer=''
     else
        buffer=substr(buffer,gtsize+3)    /* buffer now starts with a new chunk */
     iterate                        /* extract next chunk, either from buffer or from sockin */
  end

/* chunk is larger then the buffer */
  toget=(gtsize+2)-length(buffer)       /* note that chunk always has appended crlf */
  if toget<3 then do            /* special case (dangling cr or crlf */
     if toget=1 then 
        incoming=incoming||left(buffer,length(buffer)-1)
     else
        incoming=incoming||buffer

     data=sockin(socket,toget,socket_timeout,verbose)
     parse var data scode '0d0a'x data 
     if scode<0 then return data   /* some kind of error */
     if scode=0 then return '-1 {b} premature close of chunked response'
     buffer=''          /* get next chunk from sockin */
     totbytes=totbytes+length(data)
     iterate
  end

/* buffer does not contain entire chunk; so sockin some more */
  incoming=incoming||buffer

/* read chunks via calls to sockin */
  do forever            /* get this "chunk" in 1000 byte increments */
    j1000=min(1000,toget)            /* get trailing crlf */
    data=sockin(socket,j1000,socket_timeout,verbose)
    parse var data scode '0d0a'x data 
    if scode<0 then return data   /* some kind of error */
    if scode=0 then return '-1 {c} premature close of chunked response'

    lendata=length(data)
    totbytes=totbytes+lendata

    toget=toget-lendata
    if toget=0 then do        /*  end of chunk  */
        incoming=incoming||left(data,lendata-2)
        buffer=''
        leave
    end 
    else do             
        incoming=incoming||data
    end


  end                   /* forever for this chunk */

/*  large resource and tfile available?  save it to a file */
   if length(incoming)>max_notfile & tfile<>'' then do
        call write_to_tfile
   end
    
end             /* get get next chunk (exit via LEAVE */


if fileon=1 then do
    astat=sock_closetfile()
    if astat<>'' then return astat   /* error if not '' */
end

if wasclosed=1 then return ''           /* closed, to stop here */
if buffer='0d0a'x then do
  wasclosed=0
  return ''        /* no trailers */
end

if right(buffer,4)='0d0a0d0a'x then do
   trailers=left(buffer,length(buffer)-4)        
   wasclosed=0
   return ''
end

trailers=buffer

do mm=1 to 50           /*50k trailer max */
  data=sockin(socket,i1000,socket_timeout,verbose)
  parse var data scode '0d0a'x data
  if scode<0 then return data   /* some kind of error */
  if scode=0 then do          /* conection closed */
    wasclosed=1
    return ''
  end
  trailers=trailers||data
  if right(trailers,4)='0d0a0d0a' then do
     trailers=left(trailers,length(trailers)-4)
     wasclosed=0
     return ''
  end
end
return '-1 trailers too long '


/*************/
/* possibly write incoming to tfile */
write_to_tfile: 
      if fileon=0 then do               /* initialize file */
          foo=stream(tfile,'c','close')
          foo=sysfiledelete(tfile)
          foo=strip(translate(stream(tfile,'c','open write')))
          if abbrev(foo,'READY')=0 then Return '-1 unable to open file: 'tfile
          fileon=1
      end
      nm=charout(tfile,incoming)
      if nm<>0 then return '-1 error writing to file '
      incoming=''
   end 
return


/*********----------------*****************/
/* non-chunked read of content
  sets INCOMING, WASCLOSED, FILEON; writes to TFILE */

sockhttp_nonchunk:

i1000=1000
Do forever              /* build up one socket call at a time */

    if len_toget<>'' then do
        i1000=min(1000,len_toget)
    end 
    data=sockin(socket,i1000,socket_timeout,verbose)

/*  parse out status and data: -1=error, 0=closed,1=data */
    parse var data scode '0d0a'x data

    if scode<0 then return data   /* some kind of error */

    if scode=0 then do          /* conection closed */
       if fileon=1 then do
          astat=sock_closetfile()
          if astat<>'' then return astat   /* error if not '' */
       end
       wasclosed=1
       return ''
    end

/* if here, still building (sockinstat=0) */

   incoming=incoming||data


   if length(incoming)>maxlen then return '-1 Maxlen exceeded 'maxlen

/*  large resource and tfile available?  save it to a file */
   if length(incoming)>max_notfile & tfile<>'' then do
      if fileon=0 then do               /* initialize file */
          foo=stream(tfile,'c','close')
          foo=sysfiledelete(tfile)
          foo=strip(translate(stream(tfile,'c','open write')))
          if abbrev(foo,'READY')=0 then Return '-1 unable to open file: 'tfile
          fileon=1
      end
      nm=charout(tfile,incoming)
      if nm<>0 then return '-1 error writing to file '
      incoming=''
   end 

   len_toget=len_toget-length(data)

   if len_toget=0 then do                 /* content-length bytes have been read */
       if fileon=1 then do
          astat=sock_closetfile()
          if astat<>'' then return astat   /* error if not '' */
       end
       wasclosed=0              /* though it might be closed on next read */
       return ''
   end 

End              /*  ------ get next sockrecv */

/* should never get here */


/* ------------------ */
/* write to tfile, and close  */
sock_closetfile:
 nm=0
 if incoming<>'' then nm=charout(tfile,incoming)
 if nm<>0 then do
    if verbose>0 then call pmprintf('UpDaemon ERROR: unable to write to file: 'tfile)
     Return '-1 unable to write to file'
 end 
 foo=stream(tfile,'c','close')        /* don't bother checking for error, but we could */
 return ''



/*******/
/* non blocking sockrecv  */
sockin:procedure   expose outfile mm0a
parse arg socket,i1000,socket_timeout,verbose

atimeout=socket_timeout
if atimeout='' then atimeout=30
if verbose='' then verbose=0

ict=0

Do While aTimeOut > 0
    res=Sockrecv(socket,'data',i1000) 
    select 

      when res<0 then do                           /* error condition, or nothing recived? */
         If errno <> 'EWOULDBLOCK' THEN  do
            if errno='ECONNRESET' then
                Return '-2 socket (sockin) closed?  '  /* yes, real crash */
            else 
               Return '-1 crashed in sockin 'errno  /* yes, real crash */
         end

/* else not a crash, but no data available (errno = EWOULDBLOCK & sockrecv returned -1)  */
         ict=ict+1
         goo=syssleep(0.25)           /* relinquish the CPU for one second, this is from REXXUTIL */

         aTimeOut = aTimeOut - 0.25 /* count down my timer */
         Iterate                    /* loop again */
      End                           /* if sockrecv = -1 */

      when res<-1 then do            /* some other error */
         Return '-1 Error in sockin 'errno  
      end 

      when res=0 then return '0'     /* signal "closed connection */

      otherwise return '1'||'0d0a'x||data

   end
        
end             /* timeout */
return '-1 timeout'             /* if here, atimeout<0 */


/***/
a_sockclose: procedure
parse arg a1,a2   /* a2 is caller code */
foo=sockclose(a1)
return foo


/********************************************/
/* open up a socket ----------- **/
/* note that gosock is returned, or -1 error message  */ 
sockhttp_opensock:procedure   expose  outfile mm0a

parse arg gosock,server,httpport

if httpport='' then do
    if abbrev(translate(server),'HTTP://')=1 then parse var server . '://' server 
    parse var server server ':' httpport
end /* do */

if datatype(gosock)='NUM' then rc = a_sockclose(gosock,13)   /* close gosock? */

server=strip(server)
httpport=strip(httpport)
if httpport="" | datatype(httpport)<>'NUM' then httpport=80
crlf = '0d0a'x

family  ='AF_INET'

rc=1
if verify(server,'1234567890.')>0 then 
   rc=sockgethostbyname(server, "serv.0")  /* get dotaddress of server */
else
  serv.0addr=strip(server)

if rc=0 then  return '-1 ERROR Unable to resolve 'server

dotserver=serv.0addr                    /* .. */
gosaddr.0family=family                  /* set up address */
gosaddr.0port  =httpport
gosaddr.0addr  =dotserver

gosock = SockSocket(family, "SOCK_STREAM", "IPPROTO_TCP")

rc = SockConnect(gosock,"gosaddr.0")
if rc<0 then do 
     return '-1   Unable to connect to 'server' 'rc
end

return gosock

