/*25 Dec 1999. Daniel Hellerstein (danielh@crosslink.net)
  The Updater Daemon -- obtains files from an updater
  "central server", and saves to a local hard drive.

 This daemon is typically invoked by RUPDATER.CMD (when RUPDATER
 is being run as an SRE-http addon). However, ambitious users 
 can run this directly from the command line, using the following
 syntax:
 
  D:\UPDATER>UpDaemon listfile userpwd destdir tcheck dodel asem backdir range 

where:

   listfile: a fully qualified file name, pointing to a "list of files in a 
            package", as may be returned UPDATER.CMD.
   userpwd: A username and password, seperated by a :
   destdir: The "root" directory -- files obtained from the central server will
            be copied to this directory, or to subdirectories under this
            root directory.
   tcheck: Time/content check. 
             If 0, get all files. 
             If 1, check the datestamp and get if remote version is older (or
             if no remote version)
             If 2, check the crc-32 and get if remote version is different (or
             if no remote version)
   dodel: If 1, then delete files in destdir that are NOT in the listfile.
   asem: A semaphore used to communicate with the daemon. For running from the
         command line, we recommend setting this to 0
   backdir: A "backup" directory.  Files that are being replaces will be 
           "backed up" to this directory. Enter a 0 to suppress this 
           (i.e.; to NOT backup files).
   range:  A range of the "listfile" entries to use. If 0, use all entries.
           Otherwise, range should have the form   i1:i2, where i1 and i2
           are integers. For example:   120:300

Uses "package listing" files, as supplied in the command line.

*/


/**** Begin User changeable parameters ****/

/* Archive old "status" reports in  REPORTS subdirectories (of the SITE\site_nickname
   subdirectory.  1 to archive, 0 to not archive */
archive_reports=0

/* allow use of rsync encoding */
candorsync=1

/* allow use of GZIP encoding */
candogzip=1

/* number of seconds to wait between socket calls. Large values
   may be  needed over very slow lines (this is NOT the total
   time required to get a file, it's the max seconds for each
   1000 byte "sockrecv" call) */
timeout=30

/* verbosity (using PMPRINTF): 0=none, 1=some, 2= more */
verbose=1

/**** End of User changeable parameters ****/

signal on error name badguys; signal on syntax name badguys
crlf = '0d0a'x

foo=time('r')

parse arg  stuff
parse var stuff aopt listfile user destdir timecheck dodelete asem backupdir arange 
parse var user auser ':' apwd

if verbose>2 then call pmprintf('UpDater-CGI Daemon called with: 'stuff)

if aopt<>'OBTAIN' then do
   say " This UpDater daemon is not meant to be run in stand-alone mode. "
   exit
end  /* Do */

call check_dlls         /* load some dlls */

/* open output file */
oo=lastpos('.',listfile)
if oo=0 then do
   outfile=listfile'.'
end
else do
   outfile=left(listfile,oo)
end /* do */

if archive_reports=1 then do
  if stream(outfile'RPT','c','query exists')>0 then do
     rptdir=filespec('d',outfile)||filespec('p',outfile)||'REPORTS'
     if dosisdir(rptdir)=0 then foo=sysmkdir(rptdir)
     bakfile=rptdir||'\'||filespec('n',outfile)
     tfile=dostempname(bakfile'???')
     if tfile<>'' then foo=dosrename(outfile'RPT',tfile)
     foo=sysfiledelete(outfile'RPT')
  end
end
else do
   foo=sysfiledelete(outfile'RPT')
end 

outfile=outfile'RPT'
foo=stream(outfile,'c','open write')
if abbrev(translate(foo),'READY')=0 then do
   call pmprintf('UpDater-cgi Daemon ERROR ('foo'): unable to open status file 'outfile)
   exit
end /* do */

call lineout outfile,'UpDater Status Report on 'date('n')||' '||time('n')
call lineout outfile

/* read the package list */
foo=strip(translate(stream(listfile,'c','open read')))
if abbrev(foo,'READY')=0 then do
   call pmprintf('UpDater-cgi Daemon ERROR: can not open 'listfile)
   exit
end 
isz=stream(listfile,'c','query size')
if (isz=0 | isz='') then do
   call pmprintf('UpDater-cgi Daemon ERROR: empty or missing 'listfile)
   exit
end 
stuff=charin(listfile,1,isz)
foo=stream(listfile,'c','close')

/* check status of destdir */
if dosisdir(destdir)=0 then do
   foo=sysmkdir2(destdir)
   if foo<>0 then do
       call pmprintf('UpDater-cgi Daemon ERROR: unable to create destdir: 'destdir
       exit
   end /* do */
end /* do */

/*3 check status of backupdir */
if backupdir<>0 & backupdir<>'' then do
  if dosisdir(backupdir)=0 then do
   foo=sysmkdir2(backupdir)
   if foo<>0 then do
       call pmprintf('UpDater-cgi Daemon ERROR: unable to create backupdir: 'backupdir)
       exit
   end /* do */
 end /* do */
end
else do
   backupdir=''
end /* do */

parse var stuff hdrline (crlf) stuff
opts.=''
do until hdrline=''
   parse var hdrline aword hdrline
   aword=strip(aword)
   parse var aword atype '=' avalue ; atype=strip(translate(atype))
   if wordpos(atype,'PACKAGE SERVER GZIP SELECTOR RSYNC')=0 then iterate
   atail='!'||atype
   opts.atail=strip(avalue)
end 

script_name=strip(opts.!selector)

/* check for required fields */
if opts.!SERVER='' then do
       call pmprintf('UpDater-cgi Daemon ERROR: no central server info in package list 'listfile)
       exit
end /* do */

if opts.!PACKAGE='' then do
       call pmprintf('UpDater-cgi Daemon ERROR: no package name in package list 'listfile)
       exit
end /* do */

call lineout outfile,'Package: 'opts.!package 
call lineout outfile,'which is on the server at: ' opts.!server

call lineout outfile,'Package list: 'listfile
call lineout outfile,'Destination Directory: 'destdir
if backupdir<>'' then call lineout outfile,'Backup Directory: 'backupdir

/* open semapore */
if asem<>0 then do
  foo=eventsem_create(asem,'P') 
  suicide=translate(asem||'_DIE')
  foo=eventsem_close(suicide)
  foo=eventsem_create(suicide)
end

call syssleep(0.2) 
/* make a list of files to GET */
mm=0
do until stuff=''
    parse var stuff aline (crlf) stuff
    aline=strip(aline) 
    if abbrev(aline,';')=1 then iterate
    if aline='' then iterate
    mm=mm+1
    parse var aline entries.mm.!date entries.mm.!entry entries.mm.!file entries.mm.!rdir entries.mm.!noce entries.mm.!crc .
end /* do */

if mm=0 then do
   call lineout outfile,' '
   call lineout outfile,"ERROR: there are no files specified. " 
   call lineout outfile,"Perhaps there is a problem with the specification of this package? "
   call lineout outfile
   if asem<>0 then do
     foo=eventsem_close(asem)
     foo=eventsem_close(suicide)
   end
   exit
end /* do */

entries.0=mm

servername=get_hostname()   /* address of this machine */

/* now get the files ! */
stat.=0
didit.=0

retrys.=0
retrying=0

len1=length(entries.0)+1

i1=1
i2=entries.0
if pos(':',arange)>0 then do
    parse  var arange j1 ':' j2
    if datatype(j1)='NUM' & datatype(j2)='NUM' then do
         i1=j1 ; i2=j2
    end /* do */
end /* do */

tocheck=1+i2-i1
call lineout outfile,'# of files to check: 'tocheck
call lineout outfile,' '

doretry:                /* jump here to retry unGETable files */

do mm0a=i1 to i2
  if retrying=0 then
     mm0=mm0a
  else
     mm0=retrys.mm0a

  if asem<>0 then do
    aa=eventsem_query(suicide)
    if aa=-187 then foo2=eventsem_create(suicide)
    if aa>0 then do
       foo2=eventsem_close(suicide)
       call lineout outfile,'UpDater DAEMON has been killed.'
       call lineout outfile
       exit
     end 
   end          /* asem<>0 */
   cmm0=right(mm0,len1)': '
   afile=entries.mm0.!file
   afilel=strip(translate(afile,'\','/'),,'\')
   rdir0=strip(strip(translate(entries.mm0.!rdir,'\','/')),,'\')
   if rdir0='' then
      tofile=destdir||'\'||afilel
   else
      tofile=destdir||'\'||rdir0||'\'||afilel
   
   todir=filespec('d',tofile)||filespec('p',tofile)
   if dosisdir(strip(todir,,'\'))=0 then do
       foo=sysmkdir2(todir)
       if foo<>0 then do
          stat.!nocreate=stat.!nocreate+1
          call lineout outfile,cmm0'ERROR could not make directory for 'tofile
          iterate
       end
   end

   backfile=''
   if backupdir<>'' then do
      backfile=backupdir'\'afilel
      backdir=filespec('d',backfile)||filespec('p',backfile)
      if dosisdir(strip(backdir,,'\'))=0 then do
        foo=sysmkdir2(backdir)
        if foo<>0 then do
          stat.!nocreate=stat.!nocreate+1
          call lineout outfile,cmm0'ERROR could not make backup directory for 'backfile
          iterate
       end
     end                /* backupdir exists*/
   end          /* backupdir specified */

   foo='!'||strip(translate(tofile))
   didit.foo=1

/* check date first ? */
   arf=sysfiletree(tofile,'tt','FT')
   oldfile=''; timecheck=strip(timecheck)
   if tt.0<>0 then do           /* compare dates */
        oldfile=tofile
        if timecheck=1 then do                  /* update newer files only */
           parse var tt.1 todate . . .
           if left(todate,2)<70 then
              todate='20'||todate
           else
              todate='19'||todate
           if todate>= entries.mm0.!date then do
                stat.!nochange=stat.!nochange+1
                call lineout outfile,cmm0'No change (timestamp check) in 'tofile
                iterate
           end /* do */
        end                      /*timecheck */
        if timecheck=2 & entries.mm0.!crc<>'' then do          /* compare crc32?*/       
           thiscrc=fileadler(oldfile)
           if strip(translate(thiscrc))=strip(traNslate(entries.mm0.!crc)) then do
                stat.!nochange=stat.!nochange+1
                call lineout outfile,cmm0'No change (crc32 check) in 'tofile
                iterate
           end /* do */
        end /* do */
   end                /* found old version */

   aentry=entries.mm0.!entry
   req='/'script_name'?package='opts.!package'&action=file&file='afile'&entry='aentry
   if retrying=1 then req=req'&no_encoding=1'

   if timecheck=3 & candoadler=1 & oldfile<>'' then do
      dmd4=rx_md4(oldfile,1)
      req=req'&md4='dmd4
   end 

   if candogzip=1 & opts.!GZIP=1 & entries.mm0.!NOCE=0 then
            req=req'&gzip=1'
   if candorsync=1 & opts.!RSYNC=1 & entries.mm0.!NOCE=0 then do
      req=req'&rsync=1'
      rsyncok=1
   end
   else do
      rsyncok=0
   end
   stuff=get_a_url(opts.!server,req,auser,apwd,oldfile)

   if stuff='' then do 
      if notmod=1 then do               /* 304 return */
         stat.!nochange=stat.!nochange+1
         call lineout outfile,cmm0'No change (hash check) in 'tofile
         if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
         iterate
      end
      if notmod=2 then do               /* 304 return */
         stat.!nochange=stat.!nochange+1
         call lineout outfile,cmm0||'No change (md4 check) in 'tofile
         if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
         iterate
      end 
 
      stat.!noget=stat.!noget+1
      call lineout outfile,cmm0'ERROR could not GET 'afile
      if retrying=0 then do
         ir=retrys.0+1
         retrys.ir=mm0
         retrys.0=ir 
      end
      iterate
   end /* do */

   if oldfile<>"" then do               /* oldfile exists? */
     if backfile='' then do             /* then,delete it */
        fii=sysfiledelete(tofile)
     end
     else do                           /* or, move or copy oldfile to backfile */
        if translate(filespec('d',tofile))=translate(filespec('d',backfile)) then do
          fii=sysfiledelete(backfile)                   /* move old to backup */
          fii=dosrename(tofile,backfile)
          if fii<>1 then do
             call lineout outfile,fii' 'backfile'  Warning: Unable to move to backup dir: 'tofile
             foo=sysfiledelete(tofile)
          end /* do */
        end
        else do                 /* can't move, try copying */
          fii=doscopy(tofile,backfile,'R')
          if fii<>0 then do
             call lineout outfile,'Warning: Unable to copy to backup dir: 'tofile
          end /* do */
          foo=sysfiledelete(tofile)
        end
     end                        /* back file specifid */
   end                  /* oldfile exists */

   foo=stream(tofile,'c','open write')
   if abbrev(translate(foo),'READY')=0 then do
       call lineout outfile,cmm0'ERROR unable to open 'tofile
       stat.!nocreate=stat.!nocreate+1
       iterate
   end /* do */
   aa=charout(tofile,stuff,1)
   bb=stream(tofile,'c','close')
   if aa<>0 then do
       call lineout outfile,cmm0'ERROR problem writing 'tofile
       fii=sysfiledelete(tofile)
       stat.!nocreate=stat.!nocreate+1
       iterate
   end /* do */


   addo=''
   if didgzip+didrsync>0 then do
          addo=' < 'orisize' with '
          if didgzip=1 then addo=addo||' GZIP'
          if didrsync=1 then addo=addo||' RSYNC'  
   end 

   if oldfile='' then do
       stat.!created=stat.!created+1
       call lineout outfile,cmm0'Creating 'tofile '('length(stuff)' bytes'||addo
       if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
   end
   else do
       call lineout outfile,cmm0'Updating 'tofile '('length(stuff)' bytes'||addo
       stat.!updated=stat.!updated+1
       if retrying=1 then retrys.0.!ok=retrys.0.!ok+1
   end
   
end

call lineout outfile,' '
call lineout outfile,'# files processed: 'tocheck
call lineout outfile,'          created: 'stat.!created
call lineout outfile,'          updated: 'stat.!updated
call lineout outfile,'      not changed: 'stat.!nochange
nerrs=stat.!nocreate+stat.!noget
call lineout outfile,'           Errors: 'nerrs
if nerrs>0 then do
  call lineout outfile,'               GET failed= 'stat.!noget
  call lineout outfile,'            write problem= 'stat.!nocreate
end /* do */
call lineout outfile,' '
deletes=0
/* remove files? */
if dodelete=1 then do
   ff=destdir'\*.*'
   aa=sysfiletree(ff,'alls','FOS')
   do mm=1 to alls.0
      a1='!'||translate(strip(alls.mm))
      if didit.a1=0 then do
         deletes=deletes+1
         foo=sysfiledelete(alls.mm)
         if foo=0 then
            call lineout outfile,' deleted 'alls.mm 
         else
            call lineout outfile,' failed to delete 'alls.mm 
      end
   end 
   call lineout outfile,' Number of (orphan) files deleted: 'deletes'<p>'
end
if retrys.0>0  & retrying=0 then do  /* retry (once) unGETable uris? */
   call lineout outfile,' '
   call lineout outfile,' ReGET 'retrys.0' unavailable files '
   call lineout outfile,' '
   retrying=1 ;tocheck=retrys.0
   i1=1  ; i2=retrys.0
   stat.=0
   signal doretry
end

/* warn client of missing info */
if retrys.0> retrys.0.!ok  then do  
   foo=write_message(logfile,0,'')
   unavail=retrys.0-retrys.0.!Ok
   call lineout outfile,' '
   call lineout outfile,"Warning: "unavail" files were not successfully obtained"
   call lineout outfile,'You should re-access this UpDater package, request the "latest version",')
   call lineout outfile,'and try again.'
   call lineout outfile,' '
end

call lineout outfile,' Total elapsed time: '||time('e')||' seconds'
call lineout outfile
if asem<>0 then do
  foo=eventsem_close(asem)
  foo=eventsem_close(suicide)
end

exit 

/*** jump here on error */
badguys:
call pmprintf("updater-cgi daemon error "Rc " at " sigl)
exit


/********************/
/* check for unzip and rsync dlls -- we don't bother checking rexxlib, etc */
check_dlls:

candoadler=1
aa=rxfuncquery('rx_md4')
if aa=1  then do
  foo=RXFuncAdd('RXRsyncLoad', 'RXRSYNC', 'RxRsyncLoad')
  if foo=0 then call rxrsyncload
  if rxfuncquery('rx_md4')=1  then do  /* double check */
     if verbose>1 then call pmprintf('UpDater-cgi Daemon warning: rxrsync not available')
     candorsync=0 ; candoadler=0
  end
end

/* see if rxgzlib is available */
if candogzip=0 then signal check_done
aa=rxfuncquery('rxgzloadfuncs')
if aa=1  then do
  foo=RXFuncAdd('RXgzLoadfuncs', 'rxgzlib', 'RxgzyncLoadfuncs')
  if foo=0 then call rxgzloadfuncs
  if rxfuncquery('rxgzdeflatestring')=1  then do  /* double check */
     if verbose>1 then call pmprintf('UpDater-cgi Daemon warning: rxrsync not available')
     candorsync=0 ; candoadler=0
  end
end


return 1


/* ------------- */
sysmkdir2:procedure
parse arg adir

adir=strip(adir,'t','\')
ff=sysmkdir(adir)
if ff=0 then return ff

/* make the tree */
f2=adir'\'
dd=filespec('d',f2)
pp=filespec('p',f2)
if pp='\' | pp='' then return -1

pp2=strip(translate(pp,' ','\'))

do mm=1 to words(pp2)
   a1=subword(pp2,1,mm)
   a1=translate(a1,'\',' ')
   dd2=dd'\'a1
   hoo=sysmkdir(dd2)
end /* do */

return hoo


/***********/
/* get a url, using socket calls */
get_a_url:procedure expose servername crlf outfile rsyncok verbose notmod retrys. timeout ,
                        didrsync didgzip orisize
parse arg server,request,username,pwd,oldfile

arsyncok=rsyncok

if abbrev(translate(server),'HTTP://')=1 then parse var server . '://' server
req0=request
parse var server server ':' httpport
server=strip(server);
if httpport='' then httpport=80
httpport=strip(httpport)
family  ='AF_INET'
iter=0
notmod=0

agin:           /* jump here if problem (and try GET without encoding */

rc=1
if verify(server,'1234567890.')>0 then 
   rc=sockgethostbyname(server, "serv.0")  /* get dotaddress of server */
else
  serv.0addr=strip(server)

if rc=0 then do
   call lineout outfile,'ERROR Unable to resolve 'server
   return ' '
end

dotserver=serv.0addr                    /* .. */

gosaddr.0family=family                  /* set up address */
gosaddr.0port  =httpport
gosaddr.0addr  =dotserver

gosock = SockSocket(family, "SOCK_STREAM", "IPPROTO_TCP")

message='GET 'request ' HTTP/1.1'crlf'HOST:'server||crlf

message=message||'Referer:UpDater_Daemon-CGI@'||servername||crlf
if upwd<>' ' then  message=message||'Authorization: '||upwd||crlf
message=message||'Connection: close' crlf

if arsyncok=1 & oldfile<>'' then do
 rsign=rsync_synopsis(oldfile)
 message=message||'Rsync-signature_updater: 'rsign||crlf
end

if username<>'' & username<>0 then do
   upwd=mk_base64(strip(username)':'strip(pwd))
   upwd='Basic 'upwd
   message=message||'Authorization: '||upwd||crlf
end

message=message||crlf
got=''
rc = SockConnect(gosock,"gosaddr.0")
if rc<0 then do; 
  call lineout outfile,'ERROR Unable to connect to 'server
  return ' '
end
rc = SockSend(gosock, message)

/* Now get the response */
sockerr=0
got=sockin(gosock,timeout)   /* time-out aware sockrecv */
if sockerr<>0 then do
     call lineout outfile,'ERROR server not responding: 'response
     return ' '
end
rc = SockClose(gosock)

findit=crlf||crlf
ifindit=pos(findit,got)
t1=substr(got,1,ifindit)

/* look for 401 return code */
parse var t1  line1 '0d0a'x t2
parse var line1 . icode .
if icode>399  then do
  call lineout outfile,'ERROR not able ('icode') to query 'server
  return ' '
end
if icode=304 then do            /* not modified */
   notmod=1
   return ''
end /* do */

/* look for content-encoding and x-updater response headers */
cencoding='';xupdater=''

/* call lineout outfile,t2  */
clen=0
do until t2=''
   parse var t2 ahdr (crlf) t2
   parse upper var ahdr aname ':' avalue
   if strip(aname)='CONTENT-ENCODING' then cencoding=strip(avalue)
   if strip(aname)='X-UPDATER' then xupdater=strip(avalue)
   if strip(aname)='CONTENT-LENGTH' then clen=strip(translate(avalue,' ','000d0a09'x))
   if strip(aname)='X-NOTMOD' then do
     otmod=strip(translate(avalue,' ','0d0a0009'x))
     if otmod=1 then do
          notmod=2
          return ''
     end
   end
end /* do */
parse upper var xupdater . 'MD4=' smd4

stuff=substr(got,ifindit+4)

orisize=length(stuff)

if orisize<>clen then do  /* error, not correct amount of bytes recieved */
   call lineout outfile,'ERROR content-length mismatch: 'clen ', actual='||orisize
   return ' '
end /* do */

/* unencode response? */
encodeok=1
didrsync=0; didgzip=0 ; 
if cencoding<>'' then do   /* unencode from right to left */
    cencoding=translate(cencoding,' ',',')
    do jj=words(cencoding) to 1 by -1
        aword=strip(word(cencoding,jj))
        select
           when aword='GZIP' then do
              stuff=do_ungzip(stuff)
              if stuff='' then do
                if verbose>1 then call pmprintf("UpDater-cgi daemon error: bad ungzip")
                encodeok=0
                leave
              end
              didgzip=1
           end
           when aword='GDIFF' then do
              if oldfile='' then do
                encodeok=0
                leave
              end /* do */
              stuff=rsync_ungdiff(oldfile,stuff,smd4)
              if stuff='' then do
                if verbose>1 then call pmprintf('UpDater-cgi daemon error: bad ungdiff ')     
                encodeok=0
                leave
              end
              didrsync=1
           end /* do */
           if aword='IDENTITY' then iterate
           otherwise do
                encodeok=0
                leave
           end
        end  /* select */
    end /* cencoding */
end /* cencoding <> '' */

if encodeok=0 then do           /* problem encoding-- retrieve without any encoding ? */
    if iter=0 then do
        areq=''
        do until request=''
            parse var request a1 '&' request ; ta1=translate(a1)
            if abbrev(ta1,'GZIP')+abbrev(ta1,'RSYNC')=0 then areq=areq||'&'||a1
        end 
        request=strip(areq,,'&') ;iter=1 ;arsyncok=0
        signal agin
    end /* do */
    else do
        call lineout outfile,'ERROR unknown encoding encountered for 'req0
    end /* do */
end /* do */

return stuff

/************/
/* create a base64 packing of a message */
mk_base64:procedure

do mm=0 to 25           /* set base 64 encoding keys */
   a.mm=d2c(65+mm)
end /* do */
do mm=26 to 51
   a.mm=d2c(97+mm-26)
end /* do */
do mm=52 to 61
   a.mm=d2c(48+mm-52)
end /* do */
a.62='+'
a.63='/'

parse arg mess
s2=x2b(c2x(mess))
ith=0
do forever
   ith=ith+1
   a1=substr(s2,1,6,0)
   ms.ith=x2d(b2x(a1))
   if length(s2)<7 then leave
   s2=substr(s2,7)
end /* do */
pint=""
do kk=1 to ith
    oi=ms.kk ; pint=pint||a.oi
end /* do */
j1=length(pint)//4
if j1<>0 then pint=pint||copies('=',4-j1)
return pint


/*************/
/* get ip address of this machine */
get_hostname:procedure
dot_addr = SockGetHostId()
rc = SockGetHostByAddr(dot_addr, "host.")
return host.name


/***********************************/
/* un gzip a file */
do_ungzip:procedure 
parse arg astring

atid=dostid()
apid=dospid()
tmpd=value('TEMP',,'os2environment')
tmpf=tmpd'\G'||strip(left(apid,3))||atid 
tmpfgz=tmpd'\G'||strip(left(apid,3))||atid||'.GZ'
foo=sysfiledelete(tmpf)
foo=sysfiledelete(tmpfgz)

oof=strip(translate(stream(tmpfgz,'c','open write')))
if abbrev(oof,'READY')=0 then return ''         /* some kind of error */

foo=charout(tmpfgz,astring,1)
oog=stream(tmpfgz,'c','close')
if foo<>0 then do
  foo=sysfiledelete(tmpf)
  foo=sysfiledelete(tmpfgz)
  return ''
end /* do */

signal on error name badgz1; signal on syntax name badgz1
isok=0
address cmd '@gzip -d -f -q ' tmpfgz
isok=1

badgz: nop
signal on error name badguys; signal on syntax name badguys
if rc<>0 then return ''
foo=sysfiledelete(tmpfgz)

oof=strip(translate(stream(tmpf,'c','open read')))
if abbrev(oof,'READY')=0 then do
  foo=stream(tmpf,'c','close')
  foo=sysfiledelete(tmpf)
  return ''         /* some kind of error */
end
sz=stream(tmpf,'c','query size')
if sz=0 | sz='' then do
  foo=stream(tmpf,'c','close')
  foo=sysfiledelete(tmpf)
  return ''         /* some kind of error */
end
aa=charin(tmpf,1,sz)
foo=stream(tmpf,'c','close')
foo=sysfiledelete(tmpf)
return aa


/*******************************************/
rsync_synopsis:procedure 

parse arg afile,nblocks

if nblocks='' then nblocks=45
if datatype(nblocks)<>'NUM' then nblocks=45
if nblocks<10 | nblocks>255 then nblocks=45   /* 255 limit on # of blocks */

if afile='' then return "ERROR no old-version file specified"

/* read "Afile" */
aa=translate(stream(afile,'c','open read'))
if  abbrev(aa,'READY')=0 then return "ERROR could not open "afile
isize=stream(afile,'c','query size')
if isize='' | isize=0 then do
    return 'ERROR 'afile " is unaccessible"
    exit
end
astuff=charin(afile,1,isize)
aa=stream(afile,'c','close')


blocksize=trunc(0.9999 + (isize/nblocks))
if blocksize<200 then do
    blocksize=200
    nblocks=trunc((isize/blocksize)+0.999)
end /* do */
ac1=d2c(blocksize)
ac1=right(ac1,4,x2c('00'))
ac1=ac1||d2c(nblocks)
iat=1
do mm=1 to nblocks
  if mm=nblocks then
     ablock=substr(astuff,iat)
  else
     ablock=substr(astuff,iat,blocksize)
  ac0=left(x2c(rx_rsync32_md4(ablock)),8)
  ac1=ac1||ac0
  iat=iat+blocksize
end
ac1=mkpack64(ac1)

return ac1

/**********************/
mkPACK64:procedure
parse arg mess

biga=xrange('A','Z')||xrange('a','z')||xrange('0','9')||'+/'

s2=x2b(c2x(mess))

nith=trunc((length(s2)/6)+.9)
cont=copies(' ',nith)
oof=""
do mm=0 to 63
      oof=oof||x2c(b2x(right('00'||x2b(d2x(mm)),6)))
end /* do */
do ith=1 to nith 
  a1=substr(s2,(ith*6)-5,6,0)
  cont=overlay(x2c(b2x(a1)),cont,ith) 
end /* do */
pint=""
pint=translate(cont,biga,oof)
j1=length(pint)//4
if j1<>0 then pint=pint||copies('=',4-j1)
return pint


/***************************************************/
/*  rsync_ungdiff: undifference, using a gdiff difference file
                   (that may have have been produced by rsync_gdiff

     status=rsync_ungdiff(oldver_file,diff_file,newfile,GDMD4,verbo)
        OLDVER_FILE -- old version of the file
        diff_file  -- the gdiff file (in a string)
        gdmd4 -- md4 of "server side" new file -- use as an error check
                 if not specified, then error check will not be done


**************************************/

rsync_ungdiff:procedure expose outfile
  parse arg afile,s1,smd4


crlf='0d0a'x

/* read "Afile" */
aa=translate(stream(afile,'c','open read'))
if  abbrev(aa,'READY')=0 then do
  call lineout outfile,'ERROR could not open 'afile
  return ''
end
iasize=stream(afile,'c','query size')
if iasize='' | iasize=0 then do
    call lineout outfile,'ERROR unable to apply rsync to 'afile
    return ''
end
astuff=charin(afile,1,iasize)
aa=stream(afile,'c','close')

/* start building, according to the gdiff format */

/* check for gdiff header (char value of 'd1ffd1ff'x, and version # i.e; '04'x)*/
parse var s1 gheader +4 foover +1  s1
ggh=x2c('d1ffd1ff')
if gheader<>ggh then do
    call lineout outfile,'ERROR in difference file: not a gdiff file: 'c2x(gheader)
    return ''
end

/* ignore version number */
bstuff=''
noted=0
do forever
   if length(s1)=0 | s1='' then do
       call lineout outfile,'ERROR in RSYNC '
       return ''
   end
   
   parse var s1 atype +1 s1
   itype=c2d(atype)
   select
      when itype=0 then leave   /* 0 signals eof */
      when itype=255 then return 'ERROR gdiff copy operation too large'
      when  itype>0 & itype < 247 then do  /* get itype bytes */
         parse var s1 ccs +(itype) s1
         bstuff=bstuff||ccs
      end /* do */
      when itype=247 then do    /* get <65k bytes */
          parse var s1 jget +2 s1
          jget=c2d(jget)

          parse var s1 ccs +(jget) s1
          bstuff=bstuff||ccs
      end /* do */
      when itype=248 then do    /* get < 2billion bytes */
          parse var s1 jget +4 s1
          jget=c2d(jget)

          parse var s1 ccs +(jget) s1
          bstuff=bstuff||ccs
      end /* do */
      otherwise do   /* copy range of bytes */
         select
           when itype=249 then parse var s1 istart +2 iget +1 s1  
           when itype=250 then parse var s1 istart +2 iget +2 s1
           when itype=251 then parse var s1 istart +2 iget +4 s1
           when itype=252 then parse var s1 istart +4 iget +1 s1
           when itype=253 then parse var s1 istart +4 iget +2 s1
           when itype=254 then parse var s1 istart +4 iget +4 s1
           otherwise return 'ERROR: impossible code in gdiff '
         end
         istart=c2d(istart); iget=c2d(iget)

         bstuff=bstuff||substr(astuff,istart,iget)
      end               /* otherwise range */
   end                  /* gdiff codes */
end                     /* parsing gdiff file */

/* compute md4 of this constructed file */
if smd4<>'' then do 
  b2md4=rx_md4(bstuff)
  if strip(translate(b2md4)) <> strip(translate(smd4)) then do
     call lineout outfile,"ERROR md4 does not match: "b2md4', 'smd4
     return ''
  end /* do */

end

return bstuff



/***********/
/* adler-32 crc of file aname */
fileadler:procedure expose candorsync
parse arg aname

if candorsync=0 then return ''

foo=stream(aname,'c','open read')
if abbrev(strip(translate(foo)),'READY')=0 then return ''

sz=stream(aname,'c','query size')
if sz='' | sz=0 then return ' '
aa=charin(aname,1,min(100000,sz))
adler=rx_adler32(aa)
igot=100000
do forever
      if igot>=sz then leave
      iget=min(100000,sz-igot)
      aa=charin(aname,,iget)
      adler=rx_adler32(aa,adler)
      igot=igot+100000
end
foo=stream(aname,'c','close')

return adler

/**************************/
/* SOCKIN: a replacement for sockrecv. 
  call as
      stuff=sockin(socket,timeout,maxlen,timeoutmess,verbose)
  where:
     socket == a socket that's been established using sockconnect
    timeout == a timeout value in seconds
    maxlen  == maximum length of message to recieve
               If not specified, then no maximum is imposed
  timeoutmess == Prefix for "error" and "timeout" message.
                 If not specified, "#SOCKIN: " is used as a prefix
                 For example: #SOCKIN: timeout " will be returned if no response
                 was recieved in timeout seconds.
  verbose  == If 1, then report status messages. 
              If not specified, then do NOT report status messages 
and
   stuff = the contents returned from the server (up to maxlen characters)
           or an error message (starting with the timeoutmess)

Note: timeout refers to maximum seconds between "sockrecv" requests. 
      It does NOT refer to total length of time required to recieve a message.
      Thus, a number of medium length delays (say, a few seconds required
      to complete each of several sockrecv requests) will NOT cause a timeout 
      (in other words, the timeout counter is reset upon each successful 
      completion of a 1000 byte sockrecv request).

*/

SOCKIN:PROCEDURE expose sockerr
PARSE Arg socket,timeout,maxlen,timeoutmess,verbose

if maxlen=0 | maxlen='' then maxlen=100000000
if timeoutmess=''  then timeoutmess='#SOCKIN:'
if timeout='' then timeout=0.1

if verbose>2 then call pmprintf("SockIn: will wait for " timeout " seconds ....")

if Sockioctl(socket,'FIONBIO',1) = -1 then /* switch to nonblocking mode */
   Return timeoutmess||'crashed in ioctl 'errno

ict=0
ok=0
incoming=''
sockerr=1

Do While TimeOut > 0

    res=Sockrecv(socket,'data',1000) 
    if res=-1 then do                           /* error condition ? */

      If errno <> 'EWOULDBLOCK' THEN  do      /* real crash ? */
           Return timeoutmess||'crashed in sockrecv 'errno /* yes */
      end
  
      /* not-fatal,no-data-available-condition:  errno = EWOULDBLOCK & sockrecv returned -1 */

      ict=ict+1
      if verbose>2 then call pmprintf('SockIn: Waiting ('ict')...')
      Call SysSleep 1              /* relinquish the CPU for one second, this is from REXXUTIL */

      TimeOut = TimeOut - 1; /* count down my timer */
      Iterate; /* loop again */
   End; /* if sockrecv = -1 */
   if res=0 then do
        ok=1 ; leave            /* got end of message, so exit this do loop*/
   end 
    if res<0 then do
         return timeoutmess||" Error in sockrecv " rc
    end
    incoming=incoming||data
    if verbose>2 then call pmprintf("SockIn: total data recieved: "length(incoming) )
    if length(incoming)>maxlen then do
       ok=2
       leave
    end 
    ict=0               /* reset counter */
End /* do while timeout > 0 */

/* here we are timed out, or got entire message */
if ok=1  then do
   If sockioctl(socket,'FIONBIO',0) = -1 then do /* switch to blocking mode */
       if verbose>2 then call pmprintf('SockIn: ioctl error on switch to blocking mode: ' errno)
   end
   sockerr=0
   return incoming                              /* success! */
end

if ok=2  then do
   if sockioctl(socket,'FIONBIO',0) = -1 then do  /* switch to blocking mode */
      if verbose>2 then call pmprintf('SockIn: ioctl error on switch to blocking mode: ' errno)
   end
   Return timeoutmess||' message length exceeds maximum of 'maxlen
end

Return timeoutmess||' timeout '




